package cz.psika.numerist;

import java.util.Locale;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import cz.psika.utils.FileChooserLocale;

/**
 * Numerology diagram generator.
 *
 * @author Tomas Psika
 */
public final class Application implements Constants
{
    /**
     * Application starts here.
     *
     * We set locale support on application start firstly. Then preferences
     * are loaded and general style of characteristics inited. Every error
     * within this code should cause aborting of application as soon as possible.
     *
     * We could create and display Swing interface components
     * after successful initialization.
     * @param args command line arguments
     */
    public static void main(String[] args)
    {
        /* Set locales */
        Locale locale = Locale.getDefault();
        Texts.init(locale);
        FileChooserLocale.init(locale.getLanguage());

        /* Load and apply preferences before application window being shown */
        Preferences prefs = Preferences.getInstance();
        prefs.fetchUserPreferences();
        
        /* Default style for new diagrams created */
        if ( !setDefaultStyle( prefs.getPreference(Preferences.PREFERENCE_DEFAULT_STYLE) ) )
            System.exit(CONFIG_FAILURE);
        
        /* Show application window */
        SwingUtilities.invokeLater( 
            new Runnable()
            {
                @Override
                public void run()
                {
                    ApplicationWindow.getInstance().setVisible(true);
                }
            }
        );
    }

    /**
     * Set default style of characteristics.
     *
     * The style will be applied for newly created diagrams implicitly.
     * This could be overrided by selecting another style from list of 
     * all available styles later or we can change it within settings.
     *
     * @param styleName name of style
     * @return <code>>true</code> on success, <code>false</code> otherwise
     */
    private static boolean setDefaultStyle(final String styleName)
    {
        /* Set style, check it's validity, show message if error occured */
        if ( !StyleManager.setCurrentStyle(styleName) )
        {
            JOptionPane.showMessageDialog(
                null,
                Texts.get("Style_initialization_failed") + " (" + styleName + ")",
                Texts.get("Error"),
                JOptionPane.ERROR_MESSAGE
            );
            
            /* This should cause quite immediate aborting of application,
               because we have no one style to apply when creating new diagrams. */
            return false;
        }
        
        return true;
    }

    /**
     * Quit application.
     */
    public static void quit()
    {
        System.exit(OK);
    }
}
