package cz.psika.numerist;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.event.MenuEvent;

import cz.psika.utils.BrowserLauncher;
import cz.psika.utils.MenuAdapter;


/**
 * Application window.
 *
 * @author Tomas Psika
 */
public class ApplicationWindow extends JFrame implements ActionListener, Constants
{
    /**
     * Single instance.
     */
    private static ApplicationWindow wndInstance = null;
    
    /**
     * Minimal window width.
     */
    private final int MINWIDTH  = 800;
    
    /**
     * Minimal window height.
     */
    private final int MINHEIGHT = 600;
    
    /**
     * Single vs. dual view mode 
     */
    private boolean dualViewModeActive;
    
    /**
     * Default thumbnail size.
     */
    public static final int THUMBNAIL_SIZE = 100;
    
    /**
     * Default thumbnail padding.
     */
    public static final int THUMBNAIL_PADDING_HORIZ = 10;
    
    /**
     * Full diagram padding.
     */
    public static final int FULLDIAGRAM_PADDING_HORIZ = 50;
    
    /**
     * Dimension of thumbnail.
     */
    private static final Dimension THUMBNAIL_DIMENSION = new Dimension(THUMBNAIL_SIZE, THUMBNAIL_SIZE);
    
    /* Colors */
    
    private static final Color contentPanelBackgroundColor  = Color.BLACK;
    private static final Color topPanelBackgroundColor      = Color.DARK_GRAY;
    private static final Color topLabelColor                = Color.YELLOW;
    private static final Color bottomLabelColor             = Color.RED;
    
    /* Bottom command panel with buttons */
    
    /**
     * Button inside bottom panel to add new entries.
     */
    private JButton     jbAdd;
    
    /**
     * Button to change view mode in runtime.
     */
    private JButton     jbChangeMode;
    
    /**
     * Button inside bottom panel to change settings.
     */
    private JButton     jbSettings;
    
    /* Labels */
    
    /**
     * Labels showing basic information (name, copy flag) about full diagrams
     * currently shown in the panel.
     *
     * In single view mode, only one label from array is used. Array has two items
     * usable when dual view mode is active.
     */
    private JLabel      topLabels[];
    
    /**
     * Label in the bottom of main content panel.
     *
     * Shows various information messages.
     */
    private JLabel      bottomLabel;
    
    /**
     * Horizontal and vertical spaces between top labels.
     */
    private static final Insets topLabelInsets;
    
    /**
     * Horizontal and vertical spaces between bottom label and surroundings.
     */
    private static final Insets bottomLabelInsets;
    
    /**
     * Horizontal and vertical spaces between diagrams.
     */
    private static final Insets diagramInsets;
    
    static
    {
        /* Initialize insets */
        topLabelInsets = new Insets(20, 20, 20, 20);
        bottomLabelInsets = new Insets(20, 0, 20, 0);
        diagramInsets = new Insets(0, 50, 0, 50);
    }
    
    /* Menu items */
    
    private JMenuItem           jmiNewFile, jmiOpenFile, jmiSave, jmiSaveAs, jmiClose, jmiExit;
    private JMenuItem           jmiAdd, jmiDeleteSelected, jmiDeleteAll;
    private JMenuItem           jmiClearDiagrams;
    private JMenuItem           jmiAbout,jmiWebsite, jmiSupportOnline;
    private JMenu               jmSort;
    private JMenuItem           jmSortAlphabetically, jmiSortAlphabeticallyByFirstName, jmiSortAlphabeticallyByLastName, 
                                jmiSortChronologically, jmiSortByNumbers;
    private JCheckBoxMenuItem   jcbmiDescendingSortOrder;
    
    /* Content panels */
    
    /**
     * Main panel showing full diagrams.
     */
    private JPanel panel;
    
    /**
     * Panel to show thumbnails.
     */
    private JPanel topPanel;

    /**
     * We sometimes need to confirm saving changes.
     */
    private static final boolean CONFIRM_SAVE               = true;
    
    /**
     * No confirmation of saving file when saving handled by user.
     */
    private static final boolean CONFIRM_SAVE_NOT_REQUIRED  = false;
    
    /**
     * We often need to change filename when saving.
     */
    private static final boolean ASSIGN_FILE                = true;
    
    /**
     * Dont change internal filename when "Save As" feature used.
     */
    private static final boolean DONT_ASSIGN_FILE           = false;
       
    /**
     * Get window.
     *
     * @return singleton instance of <code>ApplicationWindow</code>
     */
    public static ApplicationWindow getInstance()
    {
        if (wndInstance == null)
            wndInstance = new ApplicationWindow();
        
        return wndInstance;
    }
    
    /**
     * Constructor initializes all GUI elements in application window,
     * but does not make the window visible.
     *
     * Use <code>setVisible()</code> instead.
     */
    private ApplicationWindow()
    {
        /* Set window properties */
        setMinimumSize( new Dimension(MINWIDTH, MINHEIGHT) );
        setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        setBackground(contentPanelBackgroundColor);
        
        /* Set single/dual view mode selected via preferences. */
        dualViewModeActive = VIEW_MODE_SINGLE;
        if ( Preferences.getInstance().getPreference(Preferences.PREFERENCE_APPLICATION_VIEW).equalsIgnoreCase(Preferences.PREFERENCE_DUAL_VIEW) )
            dualViewModeActive = VIEW_MODE_DUAL;
        
        /* Build GUI components */
        updateWindowTitle();
        createMenu();
        buildPanels();
        pack();
        
        /* Nearly all components need updating on resize */
        addComponentListener(new ComponentAdapter()
        {
            public void componentResized(ComponentEvent event)
            {
                resizing();
            }
        });
        setResizable(true);
        
        /* Closing application window */
        addWindowListener(new WindowAdapter()
        {
            public void windowClosing(WindowEvent e)
            {
                quit();
            }
        });
    }
    
    /**
     * Builds all content panels and bottom panel with quick access buttons.
     */
    private void buildPanels()
    {
        Container c = getContentPane();
        appendContentPanels(c);
        appendBottomPanel(c);
    }
    
    /**
     * Menubar.
     */
    private void createMenu()
    {
        /* Actions shared between various GUI components. */
        final SharedActions sharedActions = SharedActions.getInstance();
        
        /* Menubar instance */
        final JMenuBar jmbMenu = new JMenuBar();
        
        /* Initializations of menu items. */
        
        /* File */
        JMenu jmFile = new JMenu(Texts.get("File"));
        jmFile.add(jmiNewFile = new JMenuItem(Texts.get("New")));
        jmiNewFile.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.CTRL_MASK));
        jmiNewFile.addActionListener(this);
        jmFile.add(jmiOpenFile = new JMenuItem(Texts.get("Open")));
        jmiOpenFile.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, ActionEvent.CTRL_MASK));
        jmiOpenFile.addActionListener(this);
        jmFile.add(jmiSave = new JMenuItem(Texts.get("Save")));
        jmiSave.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.CTRL_MASK));
        jmiSave.addActionListener(this);
        jmFile.add(jmiSaveAs = new JMenuItem(Texts.get("Save_As")));
        jmiSaveAs.addActionListener(this);
        jmFile.add(jmiClose = new JMenuItem(Texts.get("Close")));
        jmiClose.setEnabled(false);
        jmiClose.addActionListener(this);
        jmFile.add(jmiExit = new JMenuItem(Texts.get("Exit")));
        jmiExit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
        jmiExit.addActionListener(this);
        jmFile.addMenuListener( new MenuAdapter()
            {
                public void menuSelected(MenuEvent e)
                {
                    jmiClose.setEnabled( FileManager.getInstance().getCurrentFilename() != null );
                    jmiSave.setEnabled( sharedActions.actionTest(SharedActions.Actions.SAVE_FILE) );
                }
            });

        jmbMenu.add(jmFile);

        /* Entries */
        JMenu jmEntries = new JMenu(Texts.get("Entries"));
        jmEntries.add(jmiAdd = new JMenuItem(Texts.get("Add_entry")));
        jmiAdd.addActionListener(this);
        jmEntries.addSeparator();
        jmEntries.add(jmiDeleteSelected = new JMenuItem());
        jmiDeleteSelected.setAction( sharedActions.getAction(SharedActions.Actions.DELETE_SELECTED) );
        jmiDeleteSelected.setText(Texts.get("Delete_selected"));
        jmEntries.add(jmiDeleteAll = new JMenuItem());
        jmiDeleteAll.setAction( sharedActions.getAction(SharedActions.Actions.DELETE_ALL) );
        jmiDeleteAll.setText(Texts.get("Delete_all"));
        jmEntries.addSeparator();
        jmSort = new JMenu(Texts.get("Sort"));
        jmSort.add(jmSortAlphabetically = new JMenu(Texts.get("alphabetically")));
        jmSortAlphabetically.add(jmiSortAlphabeticallyByFirstName = new JMenuItem());
        jmiSortAlphabeticallyByFirstName.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.ALPHABETICALLY_BY_FIRSTNAME) );
        jmiSortAlphabeticallyByFirstName.setText(Texts.get("by_firstname"));
        jmSortAlphabetically.add(jmiSortAlphabeticallyByLastName = new JMenuItem());
        jmiSortAlphabeticallyByLastName.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.ALPHABETICALLY_BY_LASTNAME) );
        jmiSortAlphabeticallyByLastName.setText(Texts.get("by_lastname"));
        jmSort.add(jmiSortChronologically = new JMenuItem());
        jmiSortChronologically.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.CHRONOLOGICALLY) );
        jmiSortChronologically.setText(Texts.get("chronologically"));
        jmSort.add(jmiSortByNumbers = new JMenuItem());
        jmiSortByNumbers.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.BY_NUMBER_FREQS) );
        jmiSortByNumbers.setText(Texts.get("by_number_frequencies"));
        jmEntries.add(jmSort);
        jmEntries.add(jcbmiDescendingSortOrder = new JCheckBoxMenuItem());
        jcbmiDescendingSortOrder.setState(false);
        jcbmiDescendingSortOrder.setAction( sharedActions.getAction(SharedActions.Actions.CHANGE_SORT_TYPE) );
        jcbmiDescendingSortOrder.setText(Texts.get("Descending_sorting"));
        jmEntries.addMenuListener( new MenuAdapter()
            {
                /* Handle dynamic menu items */
                public void menuSelected(MenuEvent e)
                {
                    jmiDeleteSelected.setEnabled( sharedActions.actionTest(SharedActions.Actions.DELETE_SELECTED) );
                    jmiDeleteAll.setEnabled( sharedActions.actionTest(SharedActions.Actions.DELETE_ALL) );
                    jmSort.setEnabled( sharedActions.actionTest(SharedActions.Actions.SORT_ITEMS) );
                    jcbmiDescendingSortOrder.setEnabled( sharedActions.actionTest(SharedActions.Actions.SORT_ITEMS) );
                    jcbmiDescendingSortOrder.setState( !Manager.getInstance().currentSortingOrderAscending() );
                }
            });

        jmbMenu.add(jmEntries);
        
        /* View */
        JMenu jmView = new JMenu(Texts.get("View"));
        jmView.add(jmiClearDiagrams = new JMenuItem());
        jmiClearDiagrams.setAction( sharedActions.getAction(SharedActions.Actions.CLEAR_DIAGRAMS) );
        jmiClearDiagrams.setText(Texts.get("Clear_diagrams"));
        jmView.addMenuListener( new MenuAdapter()
            {
                /* Handle dynamic menu behaviour. */
                public void menuSelected(MenuEvent e)
                {
                    jmiClearDiagrams.setEnabled( sharedActions.actionTest(SharedActions.Actions.CLEAR_DIAGRAMS) );
                }
            });

        jmbMenu.add(jmView);
        
        /* Settings, currently dialog to change settings not implemented */
        JMenu jmSettings = new JMenu(Texts.get("Tools"));
        JMenuItem jmiSettings = new JMenuItem(Texts.get("Settings"));
        jmiSettings.setEnabled(false);
        jmSettings.add(jmiSettings);

        jmbMenu.add(jmSettings);
        
        /* Help */
        JMenu jmHelp = new JMenu(Texts.get("Help"));
        jmHelp.add(jmiWebsite = new JMenuItem(Texts.get("WWW_of_project")));
        jmiWebsite.addActionListener(this);
        jmHelp.add(jmiSupportOnline = new JMenuItem(Texts.get("Support_online")));
        jmiSupportOnline.addActionListener(this);
        jmHelp.add(jmiAbout = new JMenuItem(Texts.get("About")));
        jmiAbout.addActionListener(this);
        jmbMenu.add(jmHelp);
        
        /* M$ Windows-like non-standard ALT key menu mapping (default is F10), currently this mapping disabled */
        //jmbMenu.getInputMap(JMenuBar.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_ALT, KeyEvent.ALT_DOWN_MASK, false), "takeFocus");

        setJMenuBar(jmbMenu);
    }
    
    /**
     * Add panel(s) into the middle of the window and attach listeners.
     *
     * @param c container
     */
    private void appendContentPanels(final Container c)
    {
        /* Layout constraints */
        GridBagConstraints gbConstraints = new GridBagConstraints();
        
        /* Content panel show full diagrams and additional info */
        panel = new JPanel();
        panel.setOpaque(true);
        panel.setBackground(contentPanelBackgroundColor);
        panel.setLayout(new GridBagLayout());

        /* When changing single view to dual view, the size of this array 
           and all labels may not be changed and should be reused instead */
        topLabels = new JLabel[2];
        topLabels[0] = new JLabel();
        topLabels[1] = new JLabel();
        topLabels[0].setForeground(topLabelColor);
        topLabels[1].setForeground(topLabelColor);
        
        /* Top labels */
        //gbConstraints.insets = new Insets(20, 20, 20, 20);
        gbConstraints.insets = topLabelInsets;
        panel.add(topLabels[0], gbConstraints);
        if (dualViewModeActive)
            panel.add(topLabels[1], gbConstraints);

        /* Make initial empty diagrams on startup */
        gbConstraints.gridy = 1;
        //diagramInsets = gbConstraints.insets = new Insets(0, 50, 0, 50);
        gbConstraints.insets = diagramInsets;
        panel.add( EmptyDiagram.getInstance(0), gbConstraints );
        if (dualViewModeActive)
        {
            gbConstraints.gridx = 1;
            panel.add( EmptyDiagram.getInstance(1), gbConstraints);
        }

        /* Bottom label */
        bottomLabel = new JLabel();
        bottomLabel.setForeground(bottomLabelColor);
        updateLabel(null);
        gbConstraints.gridx = 0;
        gbConstraints.gridy = 2;
        gbConstraints.gridwidth = (dualViewModeActive) ? 2 : 1;
        gbConstraints.insets = bottomLabelInsets;
        panel.add(bottomLabel, gbConstraints);
        
        /* This panel shows thumbnails */
        FlowLayout tpLayout;
        topPanel = new JPanel();
        topPanel.setOpaque(true);
        topPanel.setBackground(topPanelBackgroundColor);
        tpLayout = new FlowLayout();
        tpLayout.setHgap(THUMBNAIL_PADDING_HORIZ);
        topPanel.setPreferredSize( new Dimension(getWidth(), (int) (1.15 * THUMBNAIL_SIZE)) );
        topPanel.setBorder( BorderFactory.createLoweredBevelBorder() );
        topPanel.setLayout(tpLayout);
        topPanel.addMouseWheelListener(new topPanelMouseWheelListener());
        
        c.add(topPanel, BorderLayout.NORTH);
        c.add(panel, BorderLayout.CENTER);
    }
    
    /**
     * Bottom panel with quick access buttons.
     *
     * @param c container
     */
    private void appendBottomPanel(final Container c)
    {
        JPanel jpPanel = new JPanel();
        jpPanel.setOpaque(true);
        
        /* Adding new entry */
        jpPanel.add( jbAdd = new JButton(Texts.get("Add_new_entry")));
        jbAdd.addActionListener(this);
        
        /* Runtime change of view mode */
        jpPanel.add( jbChangeMode = new JButton(Texts.get("Dual_view_mode")) );
        jbChangeMode.addActionListener(this);
        updateChangeModeButton();

        /* Quick settings, but dialog for settings currently not implemented,
           settings could be changed manually in configuration file only */
        JButton settingsButton = new JButton(Texts.get("Settings"));
        settingsButton.setEnabled(false);
        jpPanel.add(settingsButton);
        
        c.add( jpPanel, BorderLayout.SOUTH);
    }
    
    /**
     * Refresh text in labels connected with diagrams on entry change.
     */
    public void updateLabels()
    {
        int diagramsCount = getCurrentNumberOfDiagramPositions(INSIDE_MAIN_PANEL);
        for (int pos=0; pos<diagramsCount; pos++)
        {
            Diagram tDiagram = getDiagramOnPos(pos, INSIDE_MAIN_PANEL);
            String textToSet = tDiagram.isEmpty() ? null : tDiagram.getEntry().getSourceString() + (tDiagram.isCopy() ? " / " + Texts.get("copy") : "");
            topLabels[pos].setText(textToSet);
        }
    }

    /**
     * Switching view mode in runtime.
     *
     * @param enableDualViewMode <code>true</code> when dual view mode requested
     */
    public void setViewMode(final boolean enableDualViewMode)
    {
        final boolean currentlyDualViewActive = dualViewIsActive();
        
        /* If there is nothing to change */
        if ( (currentlyDualViewActive && enableDualViewMode) || (!currentlyDualViewActive && !enableDualViewMode) )
            return;
        
        /* Must always revalidate layout hereafter */
        if (enableDualViewMode)
        {
            GridBagConstraints gbConstraints = new GridBagConstraints();
            
            /* Append new label on top */
            gbConstraints.insets = topLabelInsets;
            topLabels[1].setText("");
            panel.add(topLabels[1], gbConstraints, getDiagramPositions(INSIDE_MAIN_PANEL).next());
            panel.revalidate();
            
            /* Append new empty diagram, positions of components inside main panel
               must be incremented because of previous inserting of the label */
            gbConstraints.insets = diagramInsets;
            gbConstraints.gridx = 1;
            int firstInsertPos = getDiagramPositions(INSIDE_MAIN_PANEL).next().intValue() + 1;
            final int width = panel.getComponent(firstInsertPos).getWidth();
            EmptyDiagram emptyDiagram = EmptyDiagram.getInstance(1);
            emptyDiagram.setNewSize(new Dimension(width, width));
            panel.add(emptyDiagram, gbConstraints, ++firstInsertPos);
        }
        else
        {
            /* Remove unnecessary label and diagram */
            Diagram secondDiagramToRemove = getDiagramOnPos(1, INSIDE_MAIN_PANEL);
            panel.remove(topLabels[1]);
            panel.remove(secondDiagramToRemove);
        }
        
        dualViewModeActive = enableDualViewMode;
        
        validate();
        repaint();
    }

    /**
     * Let us know in which mode application window is currently handled.
     *
     * @return <code>true</code> when dual mode is active
     */
    public boolean dualViewIsActive()
    {
        return dualViewModeActive;
    }
    
    /**
     * Find out whether there are some items inside main content panel.
     *
     * @return <code>true</code> if panel has only empty diagrams
     */
    public boolean isContentPanelEmpty()
    {
        int comps = getMaxNoOfDiagramsInPanel(INSIDE_MAIN_PANEL);
        Iterator<Integer> iterator = getDiagramPositions(INSIDE_MAIN_PANEL);
        while (comps-- > 0/* && iterator.hasNext()*/)
            if ( !((Diagram) panel.getComponent(iterator.next())).isEmpty() )
                return false;
        
        return true;
    }
    
    /**
     * Appends, deletes or replaces diagram.
     * 
     * Thumbnail flag is specified explicitly and no diagrams are deleted on appending.
     *
     * @param diagram
     * @param pos position after which we append the diagram 
     *        or the position where we replace the diagram
     * @param isThumbnail diagram is or is not thumbnail
     */
    public void setDiagram(final Diagram diagram, final int pos, final boolean isThumbnail)
    {
        setDiagram(diagram, pos, isThumbnail, false);
    }

    /**
     * Appends or replaces diagram/thumbnail.
     *
     * @param diagram
     * @param pos the position of diagram to replace or append after
     * @throws IllegalArgumentException when trying to delete diagram on the position
     */
    public void setDiagram(final Diagram diagram, final int pos)
    {
        if (diagram == null)
            throw new IllegalArgumentException("calling append/replace diagram method, but valid diagram not supplied");
        
        setDiagram(diagram, pos, diagram.isThumbnail());
    }
    
    /**
     * Appending, replacing, inserting before a position or removing of the diagram.
     *
     * Dont call this method directly often. Try to use overloads instead.
     *
     * @param diagram       diagram to display or null when want to destroy diagram on the position
     * @param pos           position where we want to position the diagram, 
     *                      pos < 0 means (-pos + 1)th position is main content area, 
     *                      else pos = position in thumbnail area
     * @param isThumbnail
     * @param insertBefore  inserting of diagram before current position, doesnt remove anything
     */
    protected void setDiagram(Diagram diagram, int pos, boolean isThumbnail, final boolean insertBefore)
    {
        boolean remove = (diagram == null) ? true : false;
        if (!isThumbnail)
            pos = getDiagramPositions(isThumbnail).next() + pos;

        /* Remove item firstly out of the updated area if we dont want to insert diagram before the position pos */
        JPanel updatedPanel = isThumbnail ? topPanel : panel;
        if ( updatedPanel.getComponentCount() > pos && !insertBefore )
            updatedPanel.remove(pos);
        
        /* Removing diagram in main content are is the same as adding empty diagram 
           on previously changed position */
        if (remove && !isThumbnail)
        {
            diagram = EmptyDiagram.getInstance(pos - getDiagramPositions(isThumbnail).next());
            remove = false;
        }
        
        if (!remove)
        {
            /* Necessary to update component size within this layout manager correctly */
            diagram.setPreferredSize( diagram.getSize() );                              
            
            if (isThumbnail)
            {
                diagram.fetchEntry().addObserver( Manager.getInstance() );
                updatedPanel.add(diagram, pos);
            }
            else
            {
                /* Ugly grid column computation here, fixme later */
                GridBagConstraints gbConstraints = new GridBagConstraints();
                gbConstraints.gridx = (pos - 1) / 2;
                gbConstraints.gridy = 1;
                gbConstraints.insets = diagramInsets;
                updatedPanel.add(diagram, gbConstraints, pos);
            }
        }

        updateLabels();
        repaint();
        validate();
    }
    
    /**
     * Removing diagrams by position in the panel.
     *
     * @param pos position
     * @param isThumbnail
     */
    public void removeDiagram(final int pos, final boolean isThumbnail)
    {
        setDiagram(null, pos, isThumbnail);
    }
    
    /**
     * Removing concrete diagram.
     *
     * Currently called where removing full diagrams only.
     *
     * @param diagram diagram to remove
     * @param isThumbnail
     */
    public void removeDiagram(final Diagram diagram, final boolean isThumbnail)
    {
        setDiagram(null, getPosOfDiagram(diagram, isThumbnail), isThumbnail);
    }

    /**
     * Get exact position of the diagram.
     *
     * Hash code comparison must be used because "equals" method is redefined in diagrams.
     *
     * @param diagram
     * @param isThumbnail
     * @return position of the diagram inside the panel
     */
    public int getPosOfDiagram(final Diagram diagram, final boolean isThumbnail)
    {
        Iterator<Integer> iterator = getDiagramPositions(isThumbnail);
        int pos = -1, tempPos = -1;
        while ( iterator.hasNext() )
        {
            if ( getDiagramOnPos(++tempPos, isThumbnail).hashCode() == diagram.hashCode() )
            {
                pos = tempPos;
                break;
            }
            iterator.next();
        }
        
        return pos;
    }
    
    /**
     * Find exact position of the specified diagram.
     *
     * @param diagram
     * @return position of the diagram of <code>null</code> if not exists
     */
    public int getPosOfDiagram(final Diagram diagram)
    {
        return getPosOfDiagram(diagram, diagram.isThumbnail());
    }

    /**
     * Returns diagram on specified position.
     *
     * @param pos position inside a panel
     * @param insideTopPanel
     * @return diagram found or <code>null</code> is not exists
     */
    protected Diagram getDiagramOnPos(final int pos, final boolean insideTopPanel)
    {
        if ( pos >= getCurrentNumberOfDiagramPositions(insideTopPanel) )
            return null;
        
        return insideTopPanel ? (Diagram) topPanel.getComponent(pos)
            : (Diagram) panel.getComponent( getDiagramPositions(INSIDE_MAIN_PANEL).next().intValue() + pos );
    }

    /**
     * Returns number of used component positions, 
     * empty diagrams inside main content panel are included.
     *
     * @param insideTopPanel 
     * @return number of components in the panel
     */
    protected int getCurrentNumberOfDiagramPositions(final boolean insideTopPanel)
    {
        return insideTopPanel ? topPanel.getComponentCount() : getMaxNoOfDiagramsInPanel(INSIDE_MAIN_PANEL);
    }
    
    /**
     * Get first free position where a new diagram could be placed carefully.
     *
     * @param insideTopPanel 
     * @return first free position inside the relevant panel or <code>-1</code> is no free position could be found
     */
    public int getFreePosition(final boolean insideTopPanel)
    {
        int num = getMaxNoOfDiagramsInPanel(insideTopPanel);
        int cnum;
       
        if (!insideTopPanel)
        {
            /* Main content area, get first position with empty diagram */
            Iterator<Integer> positions = getDiagramPositions(FULL_DIAGRAM);
            int i = -1;
            while ( positions.hasNext() )
            {
                i++;
                if ( ( (Diagram) panel.getComponent(positions.next()) ).isEmpty() )
                    return i;
            }
        }
        else
        {
            /* Looking for free position inside thumbnail area, 
               return current number of components if some free position exists */
            cnum = topPanel.getComponentCount();
            if ( num > cnum )
                return cnum;
        }
        
        /* No free position found */
        return -1;
    }

    /**
     * Get diagram dimension according current window size.
     *
     * Calling this method often could hurt performance. 
     * You could remove unnecessary allocations by caching window size.
     *
     * @param thumbnail diagram type
     * @return dimension of this type of diagram
     */
    public Dimension getDiagramDimension(final boolean thumbnail)
    {
        if (thumbnail)
            return THUMBNAIL_DIMENSION;
        
        /* Must calculate the dimension according to current window size */
        int fullDimSize = Math.min( (int) (panel.getHeight() - THUMBNAIL_SIZE - 2 * 15), (int) (panel.getWidth() / 2 - 2 * FULLDIAGRAM_PADDING_HORIZ) );
        return new Dimension( fullDimSize, fullDimSize );
    }
    
    /**
     * Get maximal number of diagrams viewable inside thumbnail or content panel.
     *
     * @param insideTopPanel
     * @return number of max diagrams inside the panel
     */
    public int getMaxNoOfDiagramsInPanel(final boolean insideTopPanel)
    {
        if (!insideTopPanel)
            return dualViewIsActive() ? 2 : 1;
        else
            return (getWidth() - THUMBNAIL_SIZE) / (THUMBNAIL_SIZE + THUMBNAIL_PADDING_HORIZ);
    }

    /**
     * Checks whether there is possibility to insert some new item inside the panel.
     *
     * @param insideTopPanel
     * @return <code>true</code> if no diagrams are not found inside the panel
     */
    public boolean areaFree(final boolean insideTopPanel)
    {
        return getFreePosition(insideTopPanel) < 0 ? false : true;
    }
    
    /**
     * Returns iterator over component positions inside main content panel where full diagrams should be placed
     * Because currently only dual vs. single view is supported, this iterator will provide one or two items.
     * 
     * We use this method to make finding positions inside layout fast, but it has quite great disadvantage, 
     * it's not dynamic. So be careful when using this when adding or removing components inside main content panel.
     *
     * @return iterator over all diagram positions
     * @see #appendContentPanels(Container c)
     */
    private Iterator<Integer> getDiagramPositions(final boolean inTopPanel)
    {
        ArrayList<Integer> positions = new ArrayList<Integer>();
        if (inTopPanel)
        {
            int thumbnailsCount = getCurrentNumberOfDiagramPositions(THUMBNAIL);
            for (int i=0; i<thumbnailsCount; i++)
                positions.add(i);
        }
        else
        {
            if ( dualViewIsActive() )
            {
                /* First diagram position, two labels preceding inside GridBagLayout */
                positions.add(2);
                positions.add(3);
            }
            else
            {
                /* One label preceding */
                positions.add(1);
            }
        }

        return positions.iterator();
    }
    
    /**
     * Delete all diagrams.
     */
    public void deleteAll()
    {
        clearDiagrams();
        clearThumbnails();
    }
    
    /**
     * Clears diagrams in content area.
     */
    public void clearDiagrams()
    {
        for (int i=0; i<getMaxNoOfDiagramsInPanel(INSIDE_MAIN_PANEL); i++)
            setDiagram(EmptyDiagram.getInstance(i), i, INSIDE_MAIN_PANEL);
    }

    /**
     * Clears diagram thumbnails.
     *
     * @return first component's entry if any, else <code>null</code>
     */
    public Item clearThumbnails()
    {
        int nThumbnails = topPanel.getComponentCount();
        Entry first = (nThumbnails > 0) ? ((Diagram) topPanel.getComponent(0)).fetchEntry() : null;
        for (int i=0; i<nThumbnails; i++)
            removeDiagram(0, INSIDE_TOP_PANEL);
        return first;
    }

    /**
     * Handle window resizing events.
     *
     * Number of thumbnails could change and all diagram elements inside main content area
     * are going to be repainted and/or revalidated.
     */
    private void resizing()
    {
        /* Components inside content panel */
        Diagram tCanvas;
        Dimension newDimension;
        Iterator<Integer> positions = getDiagramPositions(FULL_DIAGRAM);
        while ( positions.hasNext() )
        {
            newDimension = getDiagramDimension(FULL_DIAGRAM);
            tCanvas = (Diagram) panel.getComponent( positions.next() );
            tCanvas.setNewSize( newDimension );
            tCanvas.setPreferredSize( newDimension );            
        }
        panel.validate();

        /* Check thumbnails */
        if ( topPanel.getComponentCount() > getMaxNoOfDiagramsInPanel(INSIDE_TOP_PANEL) )
        {
            /* Many thumbnails, need to remove some of them */
            while ( topPanel.getComponentCount() > getMaxNoOfDiagramsInPanel(INSIDE_TOP_PANEL) )
                removeDiagram( topPanel.getComponentCount() - 1, INSIDE_TOP_PANEL);
        }
        else
        {
            /* Less thumbnails than possible to display, must add if possible */
            Manager.getInstance().refreshThumbnails();
        }
    }
    
    /**
     * Quit application.
     *
     * We need to save changes into current file is checked, then disposes window
     * and call <code>quit()</code> method of <code>Application</code> instance at last.
     */
    private void quit()
    {
        /* Quit application, check saving file changes firstly */
        if ( FileManager.getInstance().fileModified() )
            if ( !saveDialog(CONFIRM_SAVE, ASSIGN_FILE) )
                return;

        /* Release window sources */
        dispose();
        
        /* Application exit */
        Application.quit();
    }
    
    /**
     * Update style of the button which invokes changes of view mode in runtime.
     */
    private void updateChangeModeButton()
    {
        final boolean isDualView = dualViewIsActive();
        final Color jButtonDefaultForegroundColor = UIManager.getColor("Button.foreground");
        final Font jButtonDefaultFont = UIManager.getFont("Button.font");
        jbChangeMode.setForeground( !isDualView ? jButtonDefaultForegroundColor : contentPanelBackgroundColor);
        jbChangeMode.setFont( !isDualView ? jButtonDefaultFont : new Font(jButtonDefaultFont.getName(), Font.BOLD, jButtonDefaultFont.getSize()) );
    }

    /**
     * Handle all actions.
     *
     * @param e event
     */
    public void actionPerformed(ActionEvent e)
    {
        Object actionSource = e.getSource();
        
        if (actionSource == jmiExit)
        {
            quit();
        }
        else if (actionSource == jmiAdd || actionSource == jbAdd )
        {
            /* Add new entry */
            new EntryDialog( null, StyleManager.findAllStyles(), StyleManager.getCurrentStyle().getName() );
        }
        else if (/*actionSource == jmiChangeMode || */actionSource == jbChangeMode)
        {
            setViewMode( !dualViewIsActive() );
            updateChangeModeButton();
        }
        else if (actionSource == jmiAbout)
        {
            /* Menu Help/About info dialog */
            JOptionPane.showMessageDialog(this,
                APPNAME.toUpperCase() + "\n"
                + Texts.get("Version") + ": " + VERSION + "\n"
                + Texts.get("License") + ": " + LICENSE + "\n"
                + Texts.get("Author") + ": " + AUTHORS + " (" + CONTACT_EMAIL + ")\n",
                Texts.get("About"),
                JOptionPane.INFORMATION_MESSAGE
            );
        }
        else if (actionSource == jmiWebsite || actionSource == jmiSupportOnline)
        {
            /* Invoke browser session with URL project */
            String URL = PROJECT_URL;
            if (actionSource == jmiSupportOnline)
                URL = URL.concat(PROJECT_SUPPORT_RELATIVE_URL);
            
            if ( !BrowserLauncher.launchURL(URL) )
            {
                JOptionPane.showMessageDialog(this,
                    Texts.get("Could_not_run_browser_with_URL") + URL,
                    APPNAME,
                    JOptionPane.WARNING_MESSAGE
                );
            }
        }
        else if (actionSource == jmiOpenFile)
        {
            String filename = null;
            JFileChooser fileChooser = new JFileChooser();
            int fileChoice;
            FileManager fManager = FileManager.getInstance();
            boolean openingSuccess = true;
            boolean openingCancelled = false;
            
            try
            {
                /* Check changes before opening new file */
                if ( fManager.fileModified() )
                    if ( !saveDialog(CONFIRM_SAVE, ASSIGN_FILE) )
                    {
                        openingCancelled = true;
                        return;
                    }

                /* Opening dialog */
                fileChoice = fileChooser.showOpenDialog(null);
                if ( fileChoice == JFileChooser.APPROVE_OPTION )
                {
                    filename = fileChooser.getSelectedFile().getAbsolutePath();
                    openingSuccess = Manager.getInstance().setEntriesFromFile(filename);
                }
                else
                {
                   openingCancelled = true; 
                }
            }
            catch (FileNotFoundException ex)
            {
                System.err.println("file '" + filename + "' was not found when loading entries");
                openingSuccess = false;
                JOptionPane.showMessageDialog(this, Texts.get("File_not_loaded"), Texts.get("Error"), JOptionPane.ERROR_MESSAGE);
            } 
            catch (IOException ex)
            {
                System.err.println("loading entries from file '" + filename + "' failed");
                openingSuccess = false;
                JOptionPane.showMessageDialog(this, Texts.get("File_loading_failed"), Texts.get("Error"), JOptionPane.ERROR_MESSAGE);
            }
            finally
            {
                if (!openingCancelled)
                {
                    if (!openingSuccess)
                    {
                        fManager.setCurrentFilename(null);
                        updateLabel( Texts.get("File_not_successfully_loaded") );
                    }
                    else
                        updateLabel( Texts.get("File_loaded") );

                    updateWindowTitle();
                }
            }
        }
        else if (actionSource == jmiSaveAs)
        {
            saveDialog(CONFIRM_SAVE_NOT_REQUIRED, (FileManager.getInstance().getCurrentFilename() != null ) ? DONT_ASSIGN_FILE : ASSIGN_FILE);
        }
        else if (actionSource == jmiSave)
        {
            FileManager fManager = FileManager.getInstance();
            if ( fManager.getCurrentFilename() == null )
                saveDialog(CONFIRM_SAVE_NOT_REQUIRED, ASSIGN_FILE);
            else
            {
                try
                {
                    fManager.saveFile();
                    updateLabel( Texts.get("File_saved") );
                }
                catch (IOException ex)
                {
                    saveError( fManager.getCurrentFilename() );
                }
            }
        }
        else if (actionSource == jmiClose || actionSource == jmiNewFile)
        {
            FileManager fManager = FileManager.getInstance();

            if ( fManager.fileModified() )
            {
                if (fManager.getCurrentFilename() == null)
                {
                    if ( !saveDialog(CONFIRM_SAVE, ASSIGN_FILE) )
                        return;
                }
                else
                {
                    try
                    {
                        int saveOption = JOptionPane.CANCEL_OPTION;
                        if ( (saveOption = JOptionPane.showConfirmDialog(this, Texts.get("Save_modified_file?"), 
                                Texts.get("Question"), JOptionPane.YES_NO_CANCEL_OPTION)) == JOptionPane.CANCEL_OPTION )
                            return;
                        if ( saveOption == JOptionPane.OK_OPTION )
                        {
                            fManager.saveFile();
                            updateLabel( Texts.get("File_saved") );
                        }
                    }
                    catch (IOException ex)
                    {
                        saveError( fManager.getCurrentFilename() );
                    }
                }
            }
            
            /* Delete all diagrams and clear active filename */
            Manager.getInstance().deleteAll();
            fManager.setCurrentFilename(actionSource == jmiNewFile ? Texts.get("unnamed") : null);
            fManager.setFileModifiedFlag(FILE_NOT_MODIFIED);
            updateWindowTitle();
        }
    }

    /**
     * Save confirmation dialog.
     *
     * File is going to be saved only when user confirms first questions and saves new file
     *
     * @param confirmation need to confirm saving
     * @param assign change current file on <code>true</code>
     * @return <code>true</code> on file successfully saved, <code>false</code> otherwise or on error
     */
    private boolean saveDialog(boolean confirmation, boolean assign)
    {
        JFileChooser fileChooser;
        String filename = null;
        int saveOption = JOptionPane.CANCEL_OPTION;
        int fileChoice;
        FileManager fManager = FileManager.getInstance();
        
        try
        {
            if (confirmation)
            {
                if ( (saveOption = JOptionPane.showConfirmDialog(this, Texts.get("Save_modified_file?"), 
                        Texts.get("Question"), JOptionPane.YES_NO_CANCEL_OPTION)) == JOptionPane.CANCEL_OPTION )
                    return false;
            }
            
            if ( saveOption == JOptionPane.OK_OPTION || !confirmation )
            {
                fileChooser = new JFileChooser();
                fileChoice = fileChooser.showSaveDialog(null);
                if ( fileChoice == JFileChooser.APPROVE_OPTION )
                {
                    filename = fileChooser.getSelectedFile().getAbsolutePath();
                    fManager.getInstance().saveFileAs(filename);
                    updateLabel( Texts.get("File_saved") );
                    if (assign)
                    {
                        fManager.getInstance().setCurrentFilename(filename);
                        fManager.setFileModifiedFlag(FILE_NOT_MODIFIED);
                    }
                }
                else
                    return false;
            }
            
            updateWindowTitle();
            return true;
        }
        catch (IOException ex)
        {
            saveError(filename);
        }

        return false;
    }
    
    /**
     * Alerts about error while saving file.
     *
     * @param filename
     */
    private void saveError(final String filename)
    {
        System.err.println("could not save file '" + filename + "'");
        updateLabel( Texts.get("File_not_successfully_saved") );
        JOptionPane.showMessageDialog(this, Texts.get("File_saving_failed"), Texts.get("Error"), JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Thumbnails are rotated when mouse wheel event occurs.
     */
    private class topPanelMouseWheelListener implements MouseWheelListener
    {
        public void mouseWheelMoved(MouseWheelEvent e)
        {
            if ( topPanel.getComponentCount() == getMaxNoOfDiagramsInPanel(INSIDE_TOP_PANEL) && Manager.getInstance().getNumberOfItems() > topPanel.getComponentCount() )
                Manager.getInstance().rotateThumbnails( e.getWheelRotation() );
        }
    }

    /**
     * Updates bottom label.
     *
     * @param text string to show inside label or <code>null</code> if you want to clear label's content
     */
    public void updateLabel(final String text)
    {
        if (text == null)
        {
            bottomLabel.setText(" ");
            return;
        }
        
        bottomLabel.setText( text.length() > 0 ? text : " ");
    }

    /**
     * Update window title.
     *
     * Window title consists of the application name and some useful information.
     */
    private void updateWindowTitle()
    {
        String title = APPNAME.substring(0,1).toUpperCase() +  APPNAME.substring(1);
        String filename = FileManager.getInstance().getCurrentFilename();
        if ( filename != null )
            title+=" (" + new File(filename).getName() + ")";
        setTitle(title);
    }
}
