package cz.psika.numerist;

import java.awt.Point;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.AbstractAction;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.event.PopupMenuEvent;

import cz.psika.utils.PopupMenuAdapter;

/**
 * Diagram.
 * 
 * It supports DnD operations and handle miscellaneous mouse events.
 * Diagram changes it's behaviour according to current application context.
 *
 * @author Tomas Psika
 */
public class Diagram extends DiagramCanvas
{
    /**
     * Mouse event handler.
     */
    protected MouseInputHandler mouseInputHandler;
    
    /**
     * Drag and drop supported actions.
     */
    protected int dragSourceActions;
    
    /**
     * Drop target actions.
     */
    protected int dropTargetActions;

    /**
     * Constructor creates and shows diagram canvas and attaches mouse listeners
     * and registers support for simple DnD operations.
     *
     * @param entry entry for initialization
     * @param setThumbnail what type of diagram we need
     */
    public Diagram(final Entry entry, final boolean setThumbnail)
    {
        super( entry, ApplicationWindow.getInstance().getDiagramDimension(setThumbnail), setThumbnail );
        addMouseListener(mouseInputHandler = new MouseInputHandler());
        addMouseMotionListener(mouseInputHandler);
        addDragAndDropSupport();
    }
    
    /**
     * We need empty diagram canvas here.
     *
     * This constructor make some simple diagram that draws simple rectangle
     * filled with default color in the center. Mouse listeners and DnD support
     * must also be provided. This diagram serves as possible drop target.
     */
    protected Diagram()
    {
        super();
        
        setSize(currentSize = EmptyDiagram.DIAGRAM_DIMENSION);
        setPreferredSize(EmptyDiagram.DIAGRAM_DIMENSION);
        
        setOpaque(true);
        
        entry = null;
        
        addMouseListener(mouseInputHandler = new MouseInputHandler());
        addMouseMotionListener(mouseInputHandler);
        addDragAndDropSupport();
    }
    
    /**
     * Get diagram of concrete type.
     *
     * Dont use until you really need it.
     *
     * @param wantThumbnail type of diagram we want
     * @return new <code>Diagram</code> instance if want different one or current instance
     * @see Item#fetchDiagram(boolean wantThumbnail)
     */
    /*
    public DiagramCanvas fetchDiagram(final boolean wantThumbnail)
    {
        if ( (wantThumbnail && isThumbnail()) || (!wantThumbnail && !isThumbnail()) )
            return this;
        
        return new Diagram( getEntry(), wantThumbnail );
    }
    */
    
    /**
     * This object derived from <code>DiagramCanvas</code> class could serve as drop target and drag source too.
     * Thumbnails could not be considered as drop target.
     */
    protected void addDragAndDropSupport()
    {
        if (!isEmpty())
        {
            dragSourceActions = (isThumbnail) ? DnDConstants.ACTION_COPY : DnDConstants.ACTION_COPY_OR_MOVE;
            DragSource.getDefaultDragSource().createDefaultDragGestureRecognizer(this, dragSourceActions, mouseInputHandler);
        }
        
        dropTargetActions = (isThumbnail) ? DnDConstants.ACTION_NONE : DnDConstants.ACTION_COPY_OR_MOVE;
        new DropTarget(this, dropTargetActions, mouseInputHandler, isThumbnail ? false : true);
    }
    
    /**
     * Get attached mouse input handler.
     *
     * @return attached <code>MouseInputHandler</code> object, wrapper for listening mouse and drag and drop events 
     */
    public MouseInputHandler getMouseInputHandler()
    {
        return mouseInputHandler;
    }
    
    /**
     * Class used to handle all mouse events, drag and drop inclusive.
     */
    protected class MouseInputHandler extends MouseAdapter implements ActionListener, DragGestureListener, DragSourceListener, DropTargetListener
    {
        /**
         * Popup showed on mouse click events.
         */
        private JPopupMenu  popup;
        
        /**
         * Menu items for deleting entries.
         */
        private JMenuItem   jmiDeleteAll, jmiDeleteSelected;
        
        /**
         * Menu items for clearing diagrams inside panels.
         */
        private JMenuItem   jmiClearDiagrams, jmiClearCurrentDiagram;
        
        /**
         * Sorting menu.
         */
        private JMenu       jmSort;
        
        /**
         * Sorting submenu for sorting by default alphabet.
         */
        private JMenu       jmSortAlphabetically;
        
        /**
         * Menu items for selecting concrete types of alphabetical sorting.
         */
        private JMenuItem   jmiSortAlphabeticallyByFirstName, jmiSortAlphabeticallyByLastName, jmiSortChronologically, jmiSortByNumbers;
        
        /**
         * Menu item to change sorting order.
         */
        private JCheckBoxMenuItem jcbmiDescendingSortOrder;
        
        /**
         * Menu item to let user edit the diagram.
         */
        private JMenuItem   jmiEditDiagram;
        
        /**
         * Constructor of wrapper of mouse and DnD events.
         *
         * Initializes popups and prepare of using shareable actions.
         *
         * @see SharedActions
         */
        public MouseInputHandler()
        {
            final SharedActions sharedActions = SharedActions.getInstance();
            
            /* Create popup menu */
            popup = new JPopupMenu();
            if (isThumbnail)
            {
                popup.add(jmiDeleteSelected = new JMenuItem());
                jmiDeleteSelected.setAction( SharedActions.getAction( SharedActions.Actions.DELETE_SELECTED) );
                jmiDeleteSelected.setText( Texts.get("Delete_selected") );
                popup.add(jmiDeleteAll = new JMenuItem());
                jmiDeleteAll.setAction( SharedActions.getAction( SharedActions.Actions.DELETE_ALL) );
                jmiDeleteAll.setText( Texts.get("Delete_all") );
                popup.addSeparator();
                jmSort = new JMenu(Texts.get("Sort"));
                jmSort.add(jmSortAlphabetically = new JMenu(Texts.get("alphabetically")));
                jmSortAlphabetically.add(jmiSortAlphabeticallyByFirstName = new JMenuItem());
                jmiSortAlphabeticallyByFirstName.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.ALPHABETICALLY_BY_FIRSTNAME) );
                jmiSortAlphabeticallyByFirstName.setText(Texts.get("by_firstname"));
                jmSortAlphabetically.add(jmiSortAlphabeticallyByLastName = new JMenuItem());
                jmiSortAlphabeticallyByLastName.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.ALPHABETICALLY_BY_LASTNAME) );
                jmiSortAlphabeticallyByLastName.setText(Texts.get("by_lastname"));
                jmSort.add(jmiSortChronologically = new JMenuItem());
                jmiSortChronologically.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.CHRONOLOGICALLY) );
                jmiSortChronologically.setText(Texts.get("chronologically"));
                jmSort.add(jmiSortByNumbers = new JMenuItem());
                jmiSortByNumbers.setAction( sharedActions.getAction(SharedActions.Actions.SORT_ITEMS, Manager.SORT_TYPE.BY_NUMBER_FREQS) );
                jmiSortByNumbers.setText(Texts.get("by_number_frequencies"));
                popup.add(jmSort);
                popup.add(jcbmiDescendingSortOrder = new JCheckBoxMenuItem());
                jcbmiDescendingSortOrder.setState( !Manager.getInstance().currentSortingOrderAscending() );
                jcbmiDescendingSortOrder.setAction( SharedActions.getAction(SharedActions.Actions.CHANGE_SORT_TYPE) );
                jcbmiDescendingSortOrder.setText(Texts.get("Descending_sorting"));
            }
            else
            {
                popup.add(jmiClearDiagrams = new JMenuItem());
                jmiClearDiagrams.setAction( SharedActions.getAction(SharedActions.Actions.CLEAR_DIAGRAMS) );
                jmiClearDiagrams.setText( Texts.get("Clear_diagrams") );
                popup.add(jmiClearCurrentDiagram = new JMenuItem(Texts.get("Clear_this_diagram")));
                jmiClearCurrentDiagram.addActionListener(this);
            }
            
            /* Possibility to edit entry is still visible */
            popup.addSeparator();
            popup.add(jmiEditDiagram = new JMenuItem());
            jmiEditDiagram.setAction(new AbstractAction()
            {
                @Override
                public void actionPerformed(ActionEvent e)
                {
                    new EntryDialog( entry, StyleManager.findAllStyles(), (entry != null) ? entry.getStyle().getName() : null );
                }
            });
            jmiEditDiagram.setText( !isEmpty() ? Texts.get("Edit_diagram") : Texts.get("Add_new_entry"));
            
            /* Dynamic menu changes according to current context */
            popup.addPopupMenuListener(new PopupMenuAdapter()
            {
                @Override
                public void popupMenuWillBecomeVisible(PopupMenuEvent e)
                {
                    if (isThumbnail)
                    {
                        jmiDeleteAll.setEnabled( SharedActions.actionTest(SharedActions.Actions.DELETE_ALL) );
                        jmiDeleteSelected.setEnabled( SharedActions.actionTest(SharedActions.Actions.DELETE_SELECTED) );
                        jmSort.setEnabled( SharedActions.actionTest(SharedActions.Actions.SORT_ITEMS) );
                        jcbmiDescendingSortOrder.setEnabled( SharedActions.actionTest(SharedActions.Actions.SORT_ITEMS) );
                        jcbmiDescendingSortOrder.setState( !Manager.getInstance().currentSortingOrderAscending() );
                    }
                    else
                    {
                        jmiClearDiagrams.setEnabled( SharedActions.actionTest(SharedActions.Actions.CLEAR_DIAGRAMS) );
                        if (isEmpty())
                            jmiClearCurrentDiagram.setEnabled(false);
                    }
                }
            });
        }

        /**
         * Selecting entry by clicking on thumbnail.
         *
         * @param e mouse event
         */
        @Override
        public void mouseClicked(MouseEvent e)
        {
            if (isThumbnail)
                ((Diagram) e.getSource()).setBorder(TOGGLE_BORDER);
        }
        
        /**
         * Modifies popup window according to current context.
         * 
         * @param e mouse event
         */
        private void showPopup(MouseEvent e)
        {
            if (popup.getComponentCount() > 0)
                popup.show(e.getComponent(), e.getX(), e.getY());
        }
        
        /**
         * Handle popup menu. 
         * 
         * Useful especially for Windows which triggers popups 
         * when releasing mouse button, not when pressed like on other platforms.
         * 
         * @param e mouse event
         */
        @Override
        public void mouseReleased(MouseEvent e) 
        {
            if (e.isPopupTrigger())
                showPopup(e);
        }
        
        /**
         * Showing popup menu.
         *
         * @param e mouse event
         */
        @Override
        public void mousePressed(MouseEvent e)
        {
            if (e.isPopupTrigger())
                showPopup(e);
        }
        
        /**
         * Moving of mouse pointer invokes hightlighting of some areas in full diagram
         * and updating some labels in the application frame.
         *
         * @param e mouse event
         */
        @Override
        public void mouseMoved(MouseEvent e)
        {   
            if (!isEmpty())
            {
                /* Update bottom label */
                String label = entry.getSourceString() + " ( " + getEntry().getStyle().getName() + " )";
                ApplicationWindow.getInstance().updateLabel(label);
                
                /* Highlight grid cells inside full diagrams */
                if (!isThumbnail)
                {
                    Point p = e.getPoint();
                    for (int i=0; i<Entry.NUMBERS_COUNT; i++)
                    {
                        if (rects[i].contains(p.getX(), p.getY()))
                        {
                            if (activeRectangle == i)
                                break;

                            activeRectangle = i;
                            forceRepaint();
                            break;
                        }
                    }
                }
            }
        }
        
        /**
         * Mouse exit event.
         * 
         * @param e mouse event
         * @see #mouseMoved(MouseEvent e)
         */
        @Override
        public void mouseExited(MouseEvent e)
        {
            /* Update bottom label */
            ApplicationWindow.getInstance().updateLabel(null);
            
            /* Updating previously highlighted grid cell */
            if (!isThumbnail)
            {
                activeRectangle = DiagramCanvas.NO_RECTANGLE_ACTIVE;
                forceRepaint();
            }
        }
        
        /**
         * Non-shareable actions.
         *
         * @param e action event
         */
        @Override
        public void actionPerformed(ActionEvent e)
        {
            /* Clearing current diagram */
            if ( e.getSource().equals(jmiClearCurrentDiagram) )
                ApplicationWindow.getInstance().removeDiagram( (Diagram) ((JPopupMenu) ((JMenuItem) e.getSource()).getParent()).getInvoker(), Constants.FULL_DIAGRAM );
        }
        
        /* Drag and drop functionality */
        
        @Override
        public void mouseDragged(MouseEvent e) {}
        
        /**
         * Begin to drag object.
         *
         * @param dge drag event
         */
        @Override
        public void dragGestureRecognized(DragGestureEvent dge)
        {
            Manager.getInstance().dragDiagram(dge);
        }
        
        @Override public void dragEnter(DragSourceDragEvent dsde) {}
        @Override public void dragOver(DragSourceDragEvent dsde) {}
        @Override public void dropActionChanged(DragSourceDragEvent dsde) {}
        @Override public void dragExit(DragSourceEvent dse) {}
        @Override public void dragDropEnd(DragSourceDropEvent dsde) {}
        @Override public void dragEnter(DropTargetDragEvent dtde) {}
        @Override public void dragOver(DropTargetDragEvent dtde) {}
        @Override public void dropActionChanged(DropTargetDragEvent dtde) {}
        @Override public void dragExit(DropTargetEvent dte) {}
        
        /**
         * Begin to drop object.
         *
         * @param dtde drop event
         */
        @Override
        public void drop(DropTargetDropEvent dtde)
        {
            Manager.getInstance().dropDiagram(dtde);
        }
    }
}


