package cz.psika.numerist;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;

import java.util.HashSet;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.JComponent;

/**
 * Diagram drawing component.
 *
 * Diagram is created in connection with the <code>Entry</code> object. 
 * It is observable as every object implementing <code>Item</code> interface.
 * 
 * Implicit constructor is used to create empty diagrams.
 *
 * Drawing diagrams here is quite fast and should be usable 
 * on very slow machines (<200 MHz) too. Nevertheless many graphic optimizations 
 * could be done. Drawing diagrams for entries with quite large source strings
 * could evoke disappearing of diagram(s) inside small window. 
 * Please resize window in this case, diagram should show up.
 *
 * @author Tomas Psika
 */
public class DiagramCanvas extends JComponent implements Item, Cloneable
{
    /**
     * Current dimension.
     */
    protected Dimension currentSize;
    
    /**
     * Source entry.
     */
    protected Entry entry;
    
    /**
     * Cache for current entry's style.
     */
    protected Style style;
    
    /**
     * Default arrow angle. 
     */
    protected static final double   ARROW_TRIANGLE_ANGLE        = 65;
    
    /* Some default colors */
    protected static final Color    BACKGROUND_COLOR            = Color.BLACK;
    protected static final Color    CIRCLE_COLOR                = Color.BLUE;
    protected static final Color    PLACE_BACKGROUND_COLOR      = new Color(25,25,25);
    protected static final int      MAX_NUMBER_OF_CIRCLE_COLORS = 16;
    protected static final Color    BORDER_COLOR                = Color.RED;
    protected static final Color    HIGHLIGHT_GRID_CELL_COLOR   = new Color(26,54,68);
    protected static final Color    HIGHLIGHT_LINE_COLOR        = Color.WHITE;
    protected static final Color    INVISIBLE_COLOR             = new Color(12,12,12);
    
    /**
     * Default dimension of canvas.
     *
     * //@see Item#fetchDiagram(boolean wantThumbnail)
     */
    public static final Dimension DEFAULT_DIMENSION = new Dimension(100, 100);
    
    /* Some helper variables used when drawing lines */
    protected static final int rowLines = 1;
    protected static final int colLines = 4;
    protected static final int innerDiagonalLines = 7;
    protected static final int outerDiagonalLine = 9;
    
    /**
     * Colors for numbers loaded from style or default colors.
     */
    protected Color numberColors[];
    
    /**
     * Blue circle colors, outer circles are going to be brighter.
     */
    protected static final Color circleColors[];
    static
    {
        circleColors = new Color[MAX_NUMBER_OF_CIRCLE_COLORS];
        for (int i=0; i<MAX_NUMBER_OF_CIRCLE_COLORS; i++)
            circleColors[i] = new Color(MAX_NUMBER_OF_CIRCLE_COLORS * i, MAX_NUMBER_OF_CIRCLE_COLORS * i, 255);
    }
    
    /**
     * Line strokes.
     */
    protected BasicStroke circleLineStroke, arrowLineStroke, diagonalLineStroke;
    
    /**
     * Rectangle parts of diagram.
     */
    protected Rectangle2D rects[];
    
    /**
     * Dynamic rectangles which bound circles.
     */
    protected Rectangle2D circleRects[];
    
    /**
     * Path for drawing arrowheads.
     */
    protected GeneralPath arrowhead;

    /**
     * Reusable graphic line.
     */
    protected Line2D tmpLine;
    
    /**
     * Caching image.
     */
    protected Image dImage;
    
    /**
     * Painting error.
     */
    protected boolean paintError = false;
    
    /**
     * Flag to distinguish thumbnail and full diagram.
     */
    protected final boolean isThumbnail;
    
    /**
     * Change of border required.
     */
    protected static final boolean TOGGLE_BORDER    = true;
    
    /**
     * Suppress border toggle.
     */
    protected static final boolean NO_TOGGLE_BORDER = false;

    /**
     * Diagram observers.
     */
    protected Set<ItemObserver> observers;

    /**
     * Mouse pointer is not over some rectangle.
     */
    public static final int NO_RECTANGLE_ACTIVE = -1;
    
    /**
     * Rectangle over which mouse pointer is situated.
     *
     * These diagram rectangles invoke repainting of part of diagram,
     * hightlighting lines and circles etc.
     */
    protected int activeRectangle = NO_RECTANGLE_ACTIVE;
    
    /**
     * Empty diagram flag.
     */
    protected final boolean empty;
    
    /**
     * Implicit constructor used by child class to create empty diagrams.
     *
     * Empty diagram has no observers and no attached entry.
     */
    protected DiagramCanvas()
    {
        isThumbnail = false;
        empty = true;
    }
    
    /**
     * Constructor of diagram.
     *
     * @param entry source entry
     * @param size specified dimension of requested diagram
     * @param setThumbnail thumbnail of full diagram behaviour
     */
    public DiagramCanvas(final Entry entry, final Dimension size, final boolean setThumbnail)
    {
        super();
        
        empty = false;
        
        setOpaque(true);
        setBackground(BACKGROUND_COLOR);
        setSize(currentSize = size);
        
        numberColors = new Color[Entry.NUMBERS_COUNT];

        circleRects = new Rectangle2D.Double[Entry.NUMBERS_COUNT];
        initRects();
        
        tmpLine = new Line2D.Double();

        /* Offscreen buffer */
        dImage = null;
        
        /* Set thumbnail/full diagram flag */
        isThumbnail = setThumbnail;
        
        /* Create collection of observers */
        observers = new HashSet<ItemObserver>(2);
        
        /* Set entry, this initiates painting */
        setEntry(entry);
    }

    /**
     * Set entry. 
     * 
     * Sets entry and causes diagram repaint.
     *
     * @param newEntry entry to set
     * @throws IllegalArgumentException when called for empty diagrams
     */
    public void setEntry(final Entry newEntry)
    {
        if ( isEmpty() )
            throw new IllegalArgumentException("try to set entry for empty diagram");
        
        /* Set entry and cache copy of it's style for later faster access */
        entry = newEntry;
        style = entry.getStyle();
        
        /* Prepare to repaint */
        paintError = false;
        dImage = null;
        
        /* Set border type, dont toggle */
        setBorder(NO_TOGGLE_BORDER);
        
        /* Initialize number colors according to style of characteristics, 
           every number could have its own color */
        for (int i=0; i<Entry.NUMBERS_COUNT; i++)
            numberColors[i] = style.getColorForNumber( style.getNumberOnPos(i) );

        /* Force repaint */
        repaint();
    }
    
    /**
     * Get diagram entry.
     *
     * @return entry associated with this diagram
     */
    public Entry getEntry()
    {
        return entry;
    }
    
    /**
     * Resizing of diagram.
     *
     * @param newSize new diagram dimension
     */
    public void setNewSize(final Dimension newSize)
    {
        setSize(currentSize = newSize);
        setPreferredSize(currentSize);
        
        if (!isEmpty())
            initRects();

        forceRepaint();
    }
    
    /**
     * Some diagram could be made as copy of the other one.
     * 
     * This functionality is used for making copy in dual view mode. 
     * Original diagram and it's copy could be edited side by side.
     *
     * Diagram is considered as copy of another one if it's entry is different,
     * i.e. has not the same internal identifier.
     *
     * @return <code>true</code> if this diagram is copy of another one
     * @see Entry#isCopy()
     */
    public boolean isCopy()
    {
        return entry.isCopy();
    }
    
    /**
     * Painting diagram.
     *
     * Firstly we configure graphics and then find out whether we have
     * previously painted and fresh copy of diagram in offscreen image buffer.
     * If not, painting starts into newly created offscreen image. 
     * This mechanism saves resources and runs quickly.
     *
     * @param g <code>Graphics</code> object
     */
    @Override
    public void paintComponent(Graphics g)
    {
        /* Java 2D interface */
        Graphics2D g2d = (Graphics2D) g;
        
        super.paintComponent(g2d);
        
        int cWidth = (int) currentSize.getWidth();
        int cHeight = (int) currentSize.getHeight();
        
        /* Paint only if image buffer is not available or diagram was resized */
        if ( dImage == null || dImage.getWidth(null) != cWidth  || dImage.getHeight(null) != cHeight )
        {
            /* Route painting to offscreen image */
            GraphicsConfiguration gConf = getGraphicsConfiguration();
            dImage = gConf.createCompatibleImage(cWidth, cHeight);
            g2d = (Graphics2D) dImage.getGraphics();
            
            /* Set antialiasing, dithering and other eye-candy optimizations */
            RenderingHints quality = new RenderingHints( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
            quality.put( RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY );
            quality.put( RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE );
            quality.put( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            quality.put( RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_ENABLE);
            g2d.setRenderingHints(quality);

            /* Begin painting */
            if (!isEmpty())
            {
                /* Always draw initial 3x3 grid and numbers */
                drawGrid(g2d);
                drawNumbers(g2d);

                /* If entry is specified and computed successfully, draw all circles and arrows,
                   otherwise let user know there was an error */
                if (entry != null && entry.computed())
                {
                    drawCircles(g2d);
                    drawArrows(g2d);
                }
                else
                {
                    /* Error, entry not fine, draw red rectangle now instead */
                    paintError = true;
                    g2d.setColor(Color.RED);
                    g2d.fillRect(cWidth / 4 , cHeight / 4, cWidth / 2 , cHeight / 2);
                }
            }
            else
            {
                /* Empty diagram case, just fill smaller rectangle with a color */
                g2d.setColor(INVISIBLE_COLOR);
                g2d.fillRect(cWidth / 4 , cHeight / 4, cWidth / 2 , cHeight / 2);
            }
        }
        
        /* Draw image, previously saved one or new one */
        g.drawImage(dImage, 0, 0, null);
    }

    /**
     * Force repainting of component by deleting offscreen image buffer.
     */
    public void forceRepaint()
    {
        dImage = null;
        repaint();
    }
    
    /**
     * Draws grid lines.
     *
     * Some rectangle could be hightlighted with different background color
     * when mouse pointer over it.
     *
     * @param g <code>Graphics2D</code> object
     */
    protected void drawGrid(final Graphics2D g)
    {
        g.setColor(PLACE_BACKGROUND_COLOR);
        
        for (int i=0; i<Entry.NUMBERS_COUNT; i++)
        {
            if (i==activeRectangle)
                g.setColor(HIGHLIGHT_GRID_CELL_COLOR);
            
            g.fill(rects[i]);
            
            if (i==activeRectangle)
                g.setColor(PLACE_BACKGROUND_COLOR);
        }
    }

    /**
     * Draw all numbers/digits. 
     * 
     * Each number could have it's specific color set.
     *
     * @param g <code>Graphics2D</code> object
     */
    protected void drawNumbers(final Graphics2D g)
    {
        /* Set font */
        double fontSize = currentSize.getWidth() / 12;
        Font font = new Font("Serif", Font.PLAIN, (int) fontSize);
        g.setFont(font);
        
        /* Draw digits */
        double cSize = rects[0].getWidth();
        for (int i=0; i<Entry.NUMBERS_COUNT; i++)
        {
            /* Set color for the number */
            int number = style.getNumberOnPos(i);
            g.setColor(numberColors[i]);
            
            /* FIXME - number positions are approximated, 
               only one number digit between 1 and 9 (all inclusive) supported here */
            g.drawString( String.valueOf(number),
                    (int) (rects[i].getX() + (cSize - 5 * fontSize / 8) / 2),
                    (int) (rects[i].getY() + (cSize + 2 * fontSize / 3) / 2) );
        }
    }
    
    /**
     * Drawing circles.
     *
     * @param g <code>Graphics2D</code> object
     * @throws IllegalStateException if entry not valid
     */
    protected void drawCircles(final Graphics2D g)
    {
        /* Entry validity check */
        if (!entry.computed())
            throw new IllegalStateException("entry not computed, could not draw circles");
        
        /* Set default color and stroke params */
        g.setColor(CIRCLE_COLOR);
        circleLineStroke = new BasicStroke((float) currentSize.getWidth() / 200);
        g.setStroke(circleLineStroke);
  
        /* Number of circles to draw for the numbers */
        int posFreqs[] = new int[Entry.NUMBERS_COUNT];
        
        /* Step of rectangle inflation */
        double inflateSize = rects[0].getWidth() / 30;

        /* Loop through all positions and draw circles 
           and initialize rectangles bounding circles */
        for (int i=0; i<Entry.NUMBERS_COUNT; i++)
        {
            /* Find frequency for this number, i.e. number of circles around the digit
               and then draw circles in loop */
            if ( (posFreqs[i] = entry.getFrequency(style.getNumberOnPos(i))) > 0 )
            {
                /* Initial circle width, decreasing in next loop */
                double circleWidth = rects[0].getWidth() / 4;
                
                /* Draw colorful circles from center point to outer rectangle boundaries */
                Ellipse2D.Double circle = new Ellipse2D.Double();
                for (int c = 0; c < posFreqs[i]; c++)
                {
                    circleWidth += 2 * inflateSize;
                    circleRects[i] = new Rectangle2D.Double(
                            rects[i].getX() + ( (rects[i].getWidth() - circleWidth) / 2 ),
                            rects[i].getY() + ( (rects[i].getHeight() - circleWidth) / 2 ), 
                            circleWidth, 
                            circleWidth);
                    circle.setFrame( circleRects[i] );
                    
                    g.setColor( i==activeRectangle ? HIGHLIGHT_LINE_COLOR : (c < 16 ? circleColors[c] : Color.WHITE) );
                    g.draw(circle);
                }
            }
        }
    }
    
    /**
     * Painting arrows.
     * 
     * These arrows are painted as simple lines and arrowheads are positioned and rotated.
     *
     * @param g graphic context
     * @see #drawArrowsIf
     */
    protected void drawArrows(final Graphics2D g)
    {
        /* Size of arrowhead triagle base */
        double arrowHeadSize = circleLineStroke.getLineWidth() * 6;
        
        /* Arrowline style */
        arrowLineStroke = new BasicStroke((float) currentSize.getWidth() / 150);
        float dash[] = { (float) currentSize.getWidth() / 80 };
        diagonalLineStroke = new BasicStroke((float) currentSize.getWidth() / 400, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10, dash, 0);
        g.setStroke(arrowLineStroke);
        
        /* Prepare arrowhead triangle with center point positioned in the top left corner,
           it will be used later while rotating in specific angle and in the concrete position.
           This way is quick and easy, better than use of difficult math computations. */
        arrowhead = new GeneralPath();
        double aHeight = arrowHeadSize / 2 * Math.tan( ARROW_TRIANGLE_ANGLE * Math.PI / 180);
        arrowhead.moveTo(0,0);
        arrowhead.lineTo( -arrowHeadSize / 2, -aHeight );
        arrowhead.lineTo( arrowHeadSize / 2, -aHeight );
        
        for (int pos=0; pos<9; pos++)
        {
            /* Horizontal lines */
            if ( getCol(pos+1) > getCol(pos) )
                drawArrowsIf(g, pos, pos + 1, 270);
            
            /* Vertical lines */
            if ( (pos + 3) < 9 )
                drawArrowsIf(g, pos, pos + 3, 0);
            
            /* Diagonal lines */
            
            /* Inner diagonals */
            if ( pos == 0 )
                drawArrowsIf(g, 0, 4, 135);
            else if ( pos == 8 )
                drawArrowsIf(g, 8, 4, 315);
            else if ( pos == 2 )
                drawArrowsIf(g, 2, 4, 225);
            else if ( pos == 6 )
                drawArrowsIf(g, 6, 4, 45);
            
            /* Outer diagonals */
            if ( pos == 1 )
            {
                g.setStroke(diagonalLineStroke);
                drawArrowsIf(g, 1, 3, 225);
                drawArrowsIf(g, 1, 5, 135);
                g.setStroke(arrowLineStroke);
            }
            else if (pos == 7 )
            {
                g.setStroke(diagonalLineStroke);
                drawArrowsIf(g, 7, 3, 315);
                drawArrowsIf(g, 7, 5, 45);
                g.setStroke(arrowLineStroke);
            }
        }
    }
    
    /**
     * Drawing lines with arrow(s).
     *
     * This method and/or <code>drawArrows()</code> method should be rewritten 
     * to be more readable and maybe a bit quicker. There are some things 
     * which should be removed and made more clear. But I dont have
     * much motivation to do it, because it runs quite fast for now.
     *
     * @param g graphic context
     * @param pos1 position of first rectangle
     * @param pos2 position of second rectangle
     * @param angle angle of arrowhead and line from first rectangle's point of view
     */
    protected void drawArrowsIf(final Graphics2D g, final int pos1, final int pos2, final int angle)
    {
        /* Frequencies on current positions */
        int freq1 = entry.getFrequency( style.getNumberOnPos(pos1) );
        int freq2 = entry.getFrequency( style.getNumberOnPos(pos2) );
        
        /* Continue if it is necessary to draw an arrow, else return */
        if ( freq1 == 0 || freq2 == 0 )
            return;
        
        int row1 = getRow(pos1);
        int row2 = getRow(pos2);
        int col1 = getCol(pos1);
        int col2 = getCol(pos2);
        
        tmpLine = new Line2D.Double();
        
        double width1 = circleRects[pos1].getWidth() / 2;
        double width2 = circleRects[pos2].getWidth() / 2;
        
        double circleLineWidth = circleLineStroke.getLineWidth();
        double dCircleLineWidth = 2 * circleLineWidth;
        
        /* Beginning and finish positions for the line */
        double posFromX, posFromY;
        double posToX, posToY;
        double dOffset = Math.sqrt(2) * circleLineWidth;
        double arrowOffsetFromX = 0, arrowOffsetFromY = 0, arrowOffsetToX = 0, arrowOffsetToY = 0;
        
        /* Position of arrowheads */
        Point.Double arrowHeadPositions[] = { new Point.Double(), new Point.Double() };
        double arrowHeadAngle = angle * Math.PI / 180;
        double aX, aY, angleDirection;
        
        int diagonal = 0;
        
        /* Painting lines and arrowheads, arrowheads are painted firstly */
        if ( row1 == row2 )
        {
            /* Horizontal arrows */
            g.setColor( style.getColorForLine(rowLines + row1) );
            
            posFromX    = circleRects[pos1].getCenterX() + width1 + circleLineWidth;
            posFromY    = circleRects[pos1].getCenterY();
            posToX      = circleRects[pos2].getCenterX() - width2 - circleLineWidth;
            posToY      = circleRects[pos2].getCenterY();
            
            arrowOffsetFromX = freq1 <= freq2 ? dCircleLineWidth : 0;
            arrowOffsetToX = freq1 >= freq2 ? - dCircleLineWidth : 0;
        }
        else if ( col1 == col2 )
        {
            /* Vertical arrows */
            g.setColor( style.getColorForLine(colLines + col1) );
            
            posFromX    = circleRects[pos1].getCenterX();
            posFromY    = circleRects[pos1].getCenterY() + width1 + circleLineWidth;
            posToX      = circleRects[pos2].getCenterX();
            posToY      = circleRects[pos2].getCenterY() - width2 - circleLineWidth;
            
            arrowOffsetFromY = freq1 <= freq2 ? dCircleLineWidth : 0;
            arrowOffsetToY = freq1 >= freq2 ? - dCircleLineWidth : 0;
        }
        else
        {
            /* Diagonal arrows */
            if (pos1 % 2 == 0)
            {
                /* Inner diagonals */
                if (pos1 % 8 == 0)
                    g.setColor( style.getColorForLine(innerDiagonalLines) );
                else
                    g.setColor( style.getColorForLine(innerDiagonalLines + 1) );
            }
            else
            {
                /* Outer diagonals have always the same color */
                g.setColor( style.getColorForLine(outerDiagonalLine) );
            }
            
            /* Reverse direction of arrowheads */
            diagonal = 1;
            
            /* Direction of arrowlines between positions */
            boolean xDirection = circleRects[pos1].getX() > circleRects[pos2].getX();
            boolean yDirection = circleRects[pos1].getY() > circleRects[pos2].getY();
            
            if (freq1 <= freq2)
            {
                arrowOffsetFromX = xDirection ? -dOffset : dOffset;
                arrowOffsetFromY = yDirection ? -dOffset : dOffset;
            }
            
            if (freq1 >= freq2)
            {
                arrowOffsetToX = xDirection ? dOffset : -dOffset;
                arrowOffsetToY = yDirection ? dOffset : -dOffset;
            }
            
            /* Ugly correction hack for diagonals (FIXME) */
            int cFreq1 = freq1 - freq1 / 3;
            int cFreq2 = freq2 - freq2 / 3;
            
            posFromX    = ( col1 > col2 ) ? circleRects[pos1].getCenterX() - width1 + cFreq1 * circleLineWidth :
                circleRects[pos1].getCenterX() + width1 - cFreq1 * circleLineWidth;
            posToX      = ( col1 > col2 ) ? circleRects[pos2].getCenterX() + width2 - cFreq2 * circleLineWidth :
                circleRects[pos2].getCenterX() - width2 + cFreq2 * circleLineWidth;
            posFromY    = ( row1 < row2 ) ? circleRects[pos1].getCenterY() + width1 - cFreq1 * circleLineWidth :
                circleRects[pos1].getCenterY() - width1 + cFreq1 * circleLineWidth;
            posToY      = ( row1 < row2 ) ? circleRects[pos2].getCenterY() - width2 + cFreq2 * circleLineWidth :
                circleRects[pos2].getCenterY() + width2 - cFreq2 * circleLineWidth;
        }
        
        /* We should paint at least one arrowhead per line */
        if (freq1 <= freq2)
            arrowHeadPositions[0].setLocation(posFromX, posFromY);
        if (freq1 >= freq2)
            arrowHeadPositions[1].setLocation(posToX, posToY);
        
        /* Draw line, highlight it if mouse is over specified rectangle */
        tmpLine.setLine(posFromX + arrowOffsetFromX, posFromY + arrowOffsetFromY, posToX + arrowOffsetToX, posToY + arrowOffsetToY);
        if ( pos1==activeRectangle || pos2==activeRectangle )
            g.setColor(HIGHLIGHT_LINE_COLOR);
        g.draw(tmpLine);
        
        /* Draw arrowhead(s) */
        for (int a=0; a<2; a++)
        {
            aX = arrowHeadPositions[a].getX();
            
            /* Should we draw arrowhead here ? */
            if ( aX != 0)
            {
                aY = arrowHeadPositions[a].getY();
                angleDirection = Math.PI * (1 - a + diagonal);
                
                /* Move to the position, rotate, print arrowhead and return to previous state */
                g.translate(aX, aY);
                g.rotate(arrowHeadAngle + angleDirection);
                g.fill(arrowhead);
                g.rotate(-arrowHeadAngle - angleDirection);
                g.translate(-aX, -aY);
            }
        }
    }
    
    /**
     * Set border for component or toggle border type.
     *
     * @param toggle change border style
     */
    public void setBorder(boolean toggle)
    {
        if (!isEmpty())
        {
            boolean selected = entry.isSelected();

            /* Toggle border type */
            if (toggle)
            {
                entry.setSelection(selected = !selected);
            }
            
            setBorder( selected ? BorderFactory.createLineBorder(BORDER_COLOR) : BorderFactory.createEmptyBorder() );
        }
    }

    /**
     * Refresh diagram, hide some dynamic effects temporarily.
     */
    public void refresh()
    {
        activeRectangle = NO_RECTANGLE_ACTIVE;
        forceRepaint();
    }
    
    /**
     * Find whether this diagram is the empty diagram.
     *
     * @return <code>EmptyDiagramCanvas</code> instance must return <code>true</code> here
     */
    public boolean isEmpty()
    {
        return empty;
    }

    /**
     * Check type of diagram.
     *
     * @return <code>true</code> when this diagram is thumbnail
     */
    public boolean isThumbnail()
    {
        return isThumbnail;
    }
    
    /* Item interface */
    
    /**
     * Get this diagram of concrete type.
     *
     * Dont use until you really need it.
     *
     * @param wantThumbnail type of diagram we want
     * @return new <code>DiagramCanvas</code> instance if want different one or current instance
     * @see Item#fetchDiagram(boolean wantThumbnail)
     */
    /*
    public DiagramCanvas fetchDiagram(final boolean wantThumbnail)
    {
        if ( (wantThumbnail && isThumbnail()) || (!wantThumbnail && !isThumbnail()) )
            return this;
        else
            return new DiagramCanvas(getEntry(), DiagramCanvas.DEFAULT_DIMENSION, wantThumbnail);
    }
    */
    
    /**
     * Fetch entry of this diagram.
     *
     * @return diagram's entry
     */
    @Override
    public Entry fetchEntry()
    {
        return entry;
    }
    
    /* History */
    
    /**
     * Create history object.
     *
     * @return <code>DiagramCanvasMemento</code> object representing history status of this instance
     * @see DiagramCanvasMemento#DiagramCanvasMemento(Entry entry)
     */
    @Override
    public Memento createMemento()
    {
        return new DiagramCanvasMemento(entry);
    }
    
    /**
     * Set state of this object to previously saved status.
     *
     * We just set previous entry. Repainting of diagram canvas will be done automatically.
     *
     * @param memento <code>DiagramCanvasMemento</code> object representing history status of this instance
     * @see DiagramCanvasMemento#fetchEntry()
     */
    @Override
    public void setMemento(Memento memento)
    {
        setEntry( ((DiagramCanvasMemento) memento).fetchEntry() );
    }
    
    /* Comparing and cloning of diagrams */
    
    /**
     * Diagrams are equal when their <code>Entry</code> instances are the same
     * or their unique identifiers are the same.
     *
     * @param other other <code>DiagramCanvas</code> instance
     * @return <code>true</code> if diagrams are the same
     */
    @Override
    public boolean equals(Object other)
    {
        /* Instance equality */
        if (this == other)
            return true;
        
        /* Equality of internal identifiers */
        if ( other instanceof DiagramCanvas )
        {
            Entry entry2;
            if ( entry != null && ( (entry2 = ((DiagramCanvas) other).getEntry()) != null ) )
            {
                if (entry.getId() == entry2.getId())
                    return true;
            }
        }
        
        /* Not equal */
        return false;
    }
    
    /**
     * Cloning the diagram.
     *
     * @return new cloned <code>DiagramCanvas</code> instance
     */
    @Override
    public Object clone()
    {
        return new DiagramCanvas(getEntry(), currentSize, isThumbnail());
    }
    
    /**
     * Helper method to get row number of the position.
     *
     * These methods are called quite often and causes performance loss,
     * but it is very good to use it in code to make it more readable.
     *
     * @param position position within 3x3 matrix
     * @return order number of row
     */
    private int getRow(int position)
    {
        return (int) position/3;
    }
    
    /**
     * Helper method to get the column position.
     *
     * @param position position within 3x3 matrix
     * @return order number of column
     * @see #getRow
     */
    private int getCol(int position)
    {
        return position % 3;
    }

    /**
     * Initialize all rectangles used for rendering numbers, circles and arrowheads.
     *
     * Method called when diagram is being resized or when diagram construction started,
     * not necessary to call this within <code>paintComponent</code> repeatedly.
     */
    protected void initRects()
    {
        rects = new Rectangle2D.Double[Entry.NUMBERS_COUNT];
        
        double diagramSize = currentSize.getWidth();
        
        /* Padding */
        double padding = diagramSize / 25;
        if (padding < 1)
            padding = 1;
        
        /* Get size of grid */
        double gridRectangleSize = (diagramSize - (2 * padding)) / 3;
        
        /* Initialize all grid rectangles */
        for (int i=0; i<Entry.NUMBERS_COUNT; i++)
        {
            rects[i] = new Rectangle2D.Double(
                    getCol(i) * gridRectangleSize + getCol(i) * padding,
                    getRow(i) * gridRectangleSize + getRow(i) * padding,
                    gridRectangleSize,
                    gridRectangleSize );
        }
    }
    
    /* ItemObservable interface */
    
    /**
     * Observing diagrams not used for now.
     */
    @Override
    public void addObserver(ItemObserver observer)
    {
        //observers.add(observer);
    }

    /**
     * Observing diagrams not used for now.
     */
    @Override
    public void deleteObserver(ItemObserver observer)
    {
        //observers.remove(observer);
    }

    /**
     * Observing diagrams not used for now.
     */
    @Override
    public void notifyObservers(Constants.ItemEvents event)
    {
        /*
        for (ItemObserver observer : observers)
            observer.update(this, event);
        */
    }
}
