package cz.psika.numerist;

import java.text.CharacterIterator;
import java.text.StringCharacterIterator;

import java.util.HashSet;
import java.util.Set;

/**
 * Entry. 
 * 
 * Basic object representing each entry, it's style and some computated characteristics.
 * Object is observable and could conserve history status.
 *
 * @author Tomas Psika
 */
public class Entry implements Item, History
{
    /**
     * Unique identifier for entry used to distinguish different entries.
     */
    protected final int id;
    
    /** 
     * Input string truncated for making things bit faster.
     */
    protected String name;

    /**
     * Original entry string used in GUI.
     */
    protected String sourceName;
    
    /**
     * Computed frequencies of all numbers/digits.
     */
    protected int freqs[];
    
    /**
     * Caching number of numbers with their frequency greater than zero.
     */
    protected int numbersUsed;
    
    /**
     * Style of characteristics.
     */
    protected Style style;
    
    /**
     * Successul computation status flag.
     */
    public static final boolean COMPUTED = true;
    
    /**
     * Failed computation status flag.
     */
    public static final boolean NOT_COMPUTED = false;
    
    /**
     * State of computation.
     *
     * @see #computed()
     */
    protected boolean computed = NOT_COMPUTED;
    
    /**
     * Positive selection status flag.
     */
    public static final boolean SELECTED = true;

    /**
     * Negative selection status flag.
     */
    public static final boolean NOT_SELECTED = false;
    
    /**
     * Selection of entry, used for invoking actions later.
     */
    protected boolean selectStatus = NOT_SELECTED;
    
    /**
     * Registered observers.
     */
    protected Set<ItemObserver> observers;

    /**
     * Number of numbers/digits. 
     * 
     * We suppose 3x3 matrix for all created diagrams made from the entries,
     * so 9 digits are being used.
     */
    public static final int NUMBERS_COUNT = 9;

    /**
     * Mark this entry as copy if needed.
     */
    private boolean isCopy = false;

    /**
     * Entry construction cause initialization of string content and entry identifier.
     *
     * @param uniqueIdentifier needed to observe changes quickly
     * @param entryString string input data
     * @param entryStyle style
     */
    public Entry(final Integer uniqueIdentifier, final String entryString, final Style entryStyle)
    {
        id = uniqueIdentifier;
        name = truncateWhitespaces( sourceName = entryString );
        freqs = new int[10];
        
        /* Collection of observers */
        observers = new HashSet<ItemObserver>(2);
        
        /* Apply default style */
        setStyle( style = entryStyle );
    }
    
    /**
     * Copy constructor making copy which is not connected to original entry.
     *
     * That means the copy has its own internal identifier. Only source string
     * and style are initialized. Never call this copy constructor if not sure
     * you are doing. Original and its copy does not equal.
     *
     * @param original existing entry
     * @param newId new identifier of entry to set
     * @see #equals(Object other)
     */
    protected Entry(final Entry original, final int newId)
    {
        this( newId, original.getSourceString(), original.getStyle() );
        isCopy = true;
    }
    
    /**
     * Getting unique identifier for this entry.
     *
     * @return entry identifier
     */
    public final int getId()
    {
        return id;
    }
    
    /**
     * Returns input string when this entry was created.
     *
     * @return raw input text of this entry
     */
    public String getSourceString()
    {
        return new String(sourceName);
    }
    
    /**
     * Change string and make computations.
     *
     * @param newString assing new input string
     */
    public void setString(final String newString)
    {
        name = truncateWhitespaces(sourceName = newString);
        compute();
    }
    
    /**
     * Entries could be copied.
     *
     * Copied entries have different internal identifiers.
     *
     * @return <code>true</code> when this entry has been created as copy
     */
    public boolean isCopy()
    {
        return isCopy;
    }
    
    /**
     * Return string truncated of whitespaces.
     *
     * @param stringToTruncate
     * @return truncated string
     */
    private String truncateWhitespaces(final String stringToTruncate)
    {
        StringBuffer truncatedNameBuf = new StringBuffer(25);
        StringCharacterIterator charIterator = new StringCharacterIterator(stringToTruncate);
        char b = charIterator.first();
        while ( b != CharacterIterator.DONE )
        {
            if ( !Character.isWhitespace(b) )
                truncatedNameBuf.append(b);

            b=charIterator.next();
        }
        
        return new String(truncatedNameBuf);
    }

    /* Selections */
    
    /**
     * Mark this entry as (un)selected.
     *
     * @param select <code>true</code> to select, <code>false</code> to unselect
     */
    public void setSelection(final boolean select)
    {
        selectStatus = select;
        notifyObservers( Constants.ItemEvents.EVENT_ITEM_SELECTION );
    }
    
    /**
     * Find out if this entry is selected.
     *
     * @return <code>true</code> if entry was selected, <code>false</code> otherwise
     */
    public boolean isSelected()
    {
        return selectStatus;
    }
    
    /* Style of characteristics */
    
    /**
     * Get entry's style.
     *
     * @return copy of <code>Style</code> object or <code>null</code>
     */
    public Style getStyle()
    {
        return new Style( StyleManager.getStyle(style.getName()) );
    }

    /**
     * Set/modify style for entry and make all necessary computations.
     *
     * @param setStyle style to set
     */
    public void setStyle(final Style setStyle)
    {
        style =  new Style(setStyle);
        compute();
    }

    /* Frequencies */
    
    /**
     * Get count of occurences of the concrete number/digit.
     *
     * @param number concrete number or digit
     * @return frequency count, <code>0</code> is returned if frequency not successfully computed
     * @see #computed()
     */
    public int getFrequency(final int number)
    {
        return freqs[number];
    }
    
    /**
     * Returns count of number used.
     *
     * @return numbers/digits used, <code>0</code> is returned if computation failed
     * @see #computed()
     */
    public int numbersUsed()
    {
        return numbersUsed;
    }

    /**
     * Find out if entry was successfully computed. 
     * 
     * Computations are done on every important change in object automatically.
     *
     * Call this everytime you need some informations about computation results
     * to check their validity. If you dont, results are undefined.
     *
     * @return <code>true</code> when entry computations successfully done,
     * <code>false</code> if computations are not done or not completed 
     * or when some error occured.
     */
    public boolean computed()
    {
        return computed;
    }
    
    /**
     * Initialize computed state to <code>false</code> to let us know
     * that computations are maybe processed for now and initialize
     * members to default initial state.
     */
    protected void prepareComputation()
    {
        /* Set computation state */
        computed = NOT_COMPUTED;

        /* Initial number frequencies */
        int nCounter = NUMBERS_COUNT + 1;
        while (nCounter-- != 0)
            freqs[nCounter] = 0;
        
        /* No numbers used */
        numbersUsed = 0;
    }
    
    /**
     * Set status of computation.
     *
     * @param computationStatus <code>false</code> on computation error
     * @return computation status
     */
    protected boolean setComputationStatus(final boolean computationStatus)
    {
        /* Set state to invalid, very similar to state before computation */
        if (!computationStatus)
        {
            prepareComputation();
            return NOT_COMPUTED;
        }
            
        /* Successful computation */
        return (computed = COMPUTED);
    }
    
    /**
     * Doing computations.
     *
     * @return <code>NOT_COMPUTED</code> on computation error, otherwise <code>COMPUTED</code>.
     */
    protected boolean compute()
    {
        prepareComputation();
        
        /* Compute frequencies */
        StringCharacterIterator charIterator = new StringCharacterIterator(name);
        try 
        {
            /* Get information out of the current style and count number/digit frequencies */
            String numberAssignedForCharacter;
            for ( char b=charIterator.first(); b != CharacterIterator.DONE; b=charIterator.next() )
            {
                /* Number assigned for the character read from style of characteristics.
                   If <code>null</code> initialized here, NullPointerException causes later
                   and computation for the entry will fail. */
                numberAssignedForCharacter = style.get( String.valueOf(b) );
                
                /* Some characters are going to be ignored, but this must be specified in style file */
                if (numberAssignedForCharacter.equals("?") || numberAssignedForCharacter.equals("*"))                     
                    continue;
                
                freqs[ Integer.valueOf(numberAssignedForCharacter).intValue() ]++;
            }
            
            /* Save number of numbers/digits used */
            for (int i=0; i<NUMBERS_COUNT; i++)
            {
                if ( freqs[i] > 0 )
                    numbersUsed++;
            }
        }
        catch ( RuntimeException e )
        {
            System.err.println("entry computation for entry '" + getSourceString() + "' and style '" + style.getName() + "' failed");
            return setComputationStatus(NOT_COMPUTED);
        }
       
        return setComputationStatus(COMPUTED);
    }

    /* Item interface */
    
    /**
     * Returns diagram for this entry.
     *
     * Dont use this if you really dont know what are you doing.
     * Method implemented just to satisfy interface needs and possible later use.
     *
     * @param wantThumbnail type of diagram we want
     * @return new <code>DiagramCanvas</code> instance if want different one or current instance
     * @see Item#fetchDiagram(boolean wantThumbnail)
     */
    /*
    public DiagramCanvas fetchDiagram(final boolean wantThumbnail)
    {
        return new DiagramCanvas(this, DiagramCanvas.DEFAULT_DIMENSION, wantThumbnail);
    }
    */

    /**
     * DIRECT access to this instance
     */
    @Override
    public Entry fetchEntry()
    {
        return this;
    }
    
    /**
     * Entries are equal when they have same identifiers.
     *
     * @param other the other object to compare
     * @return <code>true</code> if entries are equal
     */
    @Override
    public boolean equals(Object other)
    {
        /* Instance equality */
        if (this == other)
            return true;
        
        /* Compare identifiers */
        if ( other instanceof Entry )
        {
            if ( ((Entry) other).getId() == getId() )
                return true;
        }
        
        return false;
    }

    /* History */
    
    /**
     * Class used to conserve history status of entry.
     * No need of special implementation different from
     * the code in <code>ItemMemento</code> class.
     */
    protected class EntryMemento extends ItemMemento
    {
        /* NO EXTENSION NEEDED */
    }

    /**
     * Creates memento.
     *
     * @return new history object
     */
    @Override
    public Memento createMemento()
    {
        EntryMemento state = new EntryMemento();
        state.setString(sourceName);
        state.copyId(id);
        return state;
    }

    /**
     * Set history status.
     *
     * @param memento history <code>EntryMemento</code> object
     */
    @Override
    public void setMemento(Memento memento)
    {
        setString( ((EntryMemento) memento).getString() );
    }

    /* Observable interface */
    
    /**
     * @see ItemObservable#addObserver(ItemObserver observer)
     */
    @Override
    public void addObserver(ItemObserver observer)
    {
        observers.add(observer);
    }

    /**
     * @see ItemObservable#deleteObserver(ItemObserver observer)
     */
    @Override
    public void deleteObserver(ItemObserver observer)
    {
        observers.remove(observer);
    }

    /**
     * @see ItemObservable#notifyObservers(Constants.ItemEvents event)
     */
    @Override
    public void notifyObservers(Constants.ItemEvents event)
    {
        for (ItemObserver observer : observers)
            observer.update(this, event);
    }
}
