package cz.psika.numerist;

import cz.psika.utils.WindowCenterer;

import java.awt.FlowLayout;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;

/**
 * Dialog to add or modify entry.
 *
 * @author Tomas Psika
 */
public final class EntryDialog extends JDialog implements ActionListener
{
    /**
     * Edited entry or a new one.
     */
    private Entry entry;
    
    /**
     * Flag when adding new entry.
     */
    boolean adding;
    
    /**
     * Buttons.
     */
    private JButton jbOK, jbCancel;
    
    /**
     * Text fields.
     */
    private JTextField jtextEntry;
    
    /**
     * Comboboxes.
     */
    private JComboBox jcmbStyle;
    
    /**
     * Checkboxes.
     */
    private JCheckBox jcbUpdateContent;
    
    /**
     * Entry dialog construction.
     *
     * @param dialogEntry entry to edit or <code>null</code> if adding new entry
     * @param styleNames all styles to show in combobox
     * @param entryStyle style to apply on entry
     */
    public EntryDialog(final Entry dialogEntry, String styleNames[], final String entryStyle)
    {
        final ApplicationWindow window = ApplicationWindow.getInstance();
        
        /* Prepare entry */
        entry = dialogEntry;
        
        /* If dialogEntry is null, we are going to add new entry */
        adding = (entry == null) ? true : false;
        
        /* Set dialog properties */
        setTitle( !adding ? Texts.get("Modify_entry") : Texts.get("Add_new_entry") );
        setLayout( new BoxLayout(getContentPane(), BoxLayout.PAGE_AXIS) );
        setResizable(false);
        setModal(true);

        /* Labels, text input, combobox with applicable styles and buttons */
        JLabel jlbEntry = new JLabel(Texts.get("String_entry") + ": ");
        JLabel jlbStyle = new JLabel(Texts.get("Entry_style") + ": ");
        jtextEntry = new JTextField( adding ? "" : entry.getSourceString(), 15);
        jcmbStyle = stylesCombo(styleNames, entryStyle);

        GridLayout gridLayout = new GridLayout(2, 2, 0, 2);
        JPanel jpPanel = new JPanel();
        jpPanel.setBorder(new EmptyBorder(10,20,10,20));
        jpPanel.setLayout(gridLayout);
        jpPanel.add(jlbEntry);
        jpPanel.add(jtextEntry);
        jpPanel.add(jlbStyle);
        jpPanel.add(jcmbStyle);

        /* Checkbox for adding full diagram into main content area is enabled only 
           if there is capability to position the diagram to the window panel */
        boolean contentAreaUpdatable = window.areaFree(Constants.INSIDE_MAIN_PANEL) || !adding;
        
        FlowLayout jpCheckboxesLayout = new FlowLayout(FlowLayout.CENTER);
        JPanel jpCheckboxes = new JPanel();
        jpCheckboxes.setLayout(jpCheckboxesLayout);
        
        if (adding)
        {
            jcbUpdateContent = new JCheckBox(Texts.get("display_in_content_area"), contentAreaUpdatable);
            jcbUpdateContent.setEnabled( contentAreaUpdatable );
            jpCheckboxes.add(jcbUpdateContent);
        }

        /* Buttons */
        jbOK = new JButton(Texts.get("OK"));
        jbOK.addActionListener(this);
        jbOK.setDefaultCapable(true);
        jbCancel = new JButton(Texts.get("Cancel"));
        jbCancel.addActionListener(this);
        
        JPanel jpBtns = new JPanel();
        jpBtns.add(jbOK);
        jpBtns.add(jbCancel);
        getRootPane().setDefaultButton(jbOK);

        /* Adding this together */
        add(jpPanel);
        add(jpCheckboxes);
        add(jpBtns);
        pack();
        
        /* We want to make this dialog centered according to main application window */
        WindowCenterer.exec(this);
        
        /* Ready, show it */
        setVisible(true);
    }
    
    /**
     * Handle button events.
     *
     * @param e event
     */
    @Override
    public void actionPerformed(ActionEvent e)
    {
        if ( e.getSource() == jbCancel )
        {
            /* Dispose dialog after pressing cancel button */
            dispose();
        }
        else if ( e.getSource() == jbOK )
        {
            /* Register new entry or edit the existing one. Could append the generated diagram into panel(s). */
            
            final Manager dManager = Manager.getInstance();
            
            /* Style of the entry */
            Style entryStyle;
            if ( (entryStyle = StyleManager.getStyle((String) jcmbStyle.getSelectedItem())) == null )
            {
                /* Style error, choose current application's default style */
                System.err.println("bad style chosen while adding/modifying entry: " + (String) jcmbStyle.getSelectedItem());
                JOptionPane.showMessageDialog(this,
                    Texts.get("bad_style_default_chosen"),
                    Texts.get("Error"),
                    JOptionPane.WARNING_MESSAGE
                );
                
                entryStyle = StyleManager.getCurrentStyle();
            }
            
            /* Register new entry, or modify input text and style of the existing one and update all currently shown diagrams */
            if (adding)
                dManager.addDiagram( entry = dManager.registerEntry(jtextEntry.getText(), entryStyle), jcbUpdateContent.isSelected(), Constants.INSIDE_TOP_PANEL );
            else
            {
                entry.setString(jtextEntry.getText());
                entry.setStyle(entryStyle);
                dManager.update(entry, Constants.ItemEvents.EVENT_ENTRY_CHANGED);
            }
            
            /* File content changed */
            FileManager.getInstance().setFileModifiedFlag(Constants.FILE_MODIFIED);
            
            /* Nothing more to do, close dialog */
            dispose();
        }
    }

    /**
     * Create combobox with list of applicable styles and preselects the selected one.
     *
     * @param styleNames
     * @param selectedStyle
     * @return generated combobox
     */
    private JComboBox stylesCombo(String styleNames[], final String selectedStyle)
    {
        JComboBox comboStyle = new JComboBox();
        for (int t=0; t<styleNames.length; t++)
        {
            comboStyle.addItem(styleNames[t]);
            if ( styleNames[t].equals(selectedStyle) )
                comboStyle.setSelectedIndex(t);
        }
        return comboStyle;
    }
}
