package cz.psika.numerist;

import cz.psika.numerist.Constants.ItemEvents;

import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DropTargetDropEvent;

import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.Vector;

import javax.swing.JOptionPane;

/**
 * Manager for diagrams and corresponding entries.
 *
 * @author Tomas Psika
 */
public final class Manager implements Constants, ItemObserver
{
    /**
     * Single diagram manager instance.
     */
    private static Manager manager = null;
    
    /**
     * Main application window.
     */
    private static ApplicationWindow window = null;
    
    /**
     * Sequence for generating unique identifiers 
     * for registered entries.
     */
    private static Integer entrySeq = 0;
    
    /**
     * Vector of all entries.
     */
    private Vector<Item> items;
    
    /**
     * Count of selected items.
     */
    private int selectedItemsCount = 0;                                         

    /* Sorting */
    
    /**
     * Type of sorting.
     * 
     * @author Tomas Psika
     */
    public static enum SORT_TYPE
    {
        CHRONOLOGICALLY,                                                        /* Chronologically, i.e. sorted by unique identifier */
        ALPHABETICALLY_BY_FIRSTNAME,
        ALPHABETICALLY_BY_LASTNAME,
        BY_NUMBER_FREQS                                                         /* By frequency of computed numbers */
    };
    
    /**
     * Sorting order.
     * 
     * @author Tomas Psika
     */
    public static enum SORT_ORDER
    {
        ASCENDING,
        DESCENDING
    };
    
    /**
     * Current sorting order.
     */
    private static SORT_ORDER currentSortOrder = SORT_ORDER.ASCENDING;
    
    /* Comparators */
    
    private static Comparator<Item> chronoComparator;
    private static Comparator<Item> alphaComparatorByFirstName;
    private static Comparator<Item> alphaComparatorByLastName;
    private static Comparator<Item> numbersComparator;
    
    static
    {
        chronoComparator = new ChronoComparator<Item>();
        alphaComparatorByFirstName = new AlphaComparatorByFirstName<Item>();
        alphaComparatorByLastName = new AlphaComparatorByLastName<Item>();
        numbersComparator = new NumbersComparator<Item>();
    }
    
    /**
     * Returns singleton instance.
     *
     * @return instance of the diagram manager
     */
    public static Manager getInstance()
    {
        if (manager == null)
        {
            if (window == null)
                window = ApplicationWindow.getInstance();
            
            manager = new Manager();
        }
        
        return manager;
    }

    /**
     * Constructor.
     */
    private Manager()
    {
        items = new Vector<Item>();
    }
    
    /**
     * Register new entry within collection.
     *
     * Created entry has it's own unique identifier and is registered 
     * inside internal <code>Vector</code> of <code>Item</code>s.
     *
     * @param entryString input text
     * @param entryStyle input style
     * @return new <code>Entry</code> object registered
     */
    public Entry registerEntry(String entryString, Style entryStyle)
    {
        Entry e = new Entry(++entrySeq, entryString, entryStyle);
        items.add(e);
        return e;
    }

    /**
     * Get iterator over all items.
     *
     * @return iterator
     */
    public Iterator<Item> itemsIterator()
    {
        return items.iterator();
    }
    
    /**
     * Get current number of items.
     *
     * @return number of items
     */
    public int getNumberOfItems()
    {
        return items.size();
    }

    /**
     * Fetch diagram for this concrete item.
     *
     * @param item
     * @param isThumbnail thumbnail flag
     * @return new diagram
     */
    public Diagram getDiagram(Item item, boolean isThumbnail)
    {
        return new Diagram( item.fetchEntry(), isThumbnail);
    }
    
    /**
     * Get diagram copy for the item.
     *
     * @param item
     * @param isThumbnail thumbnail flag
     * @return new copy of diagram
     */
    public Diagram getDiagramCopy(Item item, boolean isThumbnail)
    {
        return getDiagram( new Entry(item.fetchEntry(), ++entrySeq), isThumbnail );
    }

    /**
     * Load entries from file and set thumbnails.
     *
     * All previous entries are going to be deleted.
     *
     * @param filename
     * @return <code>false</code> on fatal error
     * @throws IOException on some i/o error
     * @throws FileNotFoundException when file not found
     */
    public boolean setEntriesFromFile(final String filename) throws FileNotFoundException, IOException
    {
        boolean status = true;
        FileManager fManager = FileManager.getInstance();
        
        try 
        {
            Vector<String> lines = fManager.loadFile(filename);
            Iterator<String> lineIterator = lines.iterator();
            
            /* Delete all previous entries registered after successful reading */
            deleteAll();
            
            /* Register new items and append to top panel as thumbnail */
            boolean panelIsFull = false;
            Style currentStyle = StyleManager.getCurrentStyle();
            while ( lineIterator.hasNext() )
            {
                Entry newEntry = registerEntry(lineIterator.next(), currentStyle);
                if (!panelIsFull)
                {
                    addDiagram(newEntry, false, true);
                    panelIsFull = window.getFreePosition(INSIDE_TOP_PANEL) < 0 ? true : false;
                }
            }
        }
        catch(OutOfMemoryError e)
        {
            /* Release all possible objects from memory at first */
            deleteAll();
            
            status = false;
            fManager.setCurrentFilename(null);
            
            System.err.println("out of memory when loading entries");
            JOptionPane.showMessageDialog(ApplicationWindow.getInstance(),
                Texts.get("Out_of_memory_too_many_entries_loaded"),
                Texts.get("Warning"),
                JOptionPane.INFORMATION_MESSAGE
            );
        }
        finally
        {
            /* We unset file modified status flag in every case */
            fManager.setFileModifiedFlag(FILE_NOT_MODIFIED);
        }

        return status;
    }

    /* Sorting & comparators */

    /**
     * Comparing items by time of creation.
     *
     * Items with higher internal identifier are considered as newer then the others.
     */
    private static class ChronoComparator<T extends Item> implements Comparator<T>
    {
        @Override
        public int compare(T o1, T o2)
        {
            int id1 = o1.fetchEntry().getId();
            int id2 = o2.fetchEntry().getId();
            if (id1 == id2)
                return 0;
            else
                return (id1 > id2) ? 1 : -1;
        }
    };
    
    /**
     * Comparing itmes by firstname alphabetically.
     */
    private static class AlphaComparatorByFirstName<T extends Item> implements Comparator<T>
    {
        @Override
        public int compare(T o1, T o2)
        {
            return o1.fetchEntry().getSourceString().compareToIgnoreCase(o2.fetchEntry().getSourceString());
        }
    };
    
    /**
     * Comparing itmes by lastname alphabetically.
     */
    private static class AlphaComparatorByLastName<T extends Item> implements Comparator<T>
    {
        @Override
        public int compare(T o1, T o2)
        {
            String name = o1.fetchEntry().getSourceString();
            String lastName1 = (name.lastIndexOf(" ") != -1) ? name.substring(name.lastIndexOf(" ") + 1) : name;
            name = o2.fetchEntry().getSourceString();
            String lastName2 = (name.lastIndexOf(" ") != -1) ? name.substring(name.lastIndexOf(" ") + 1) : name;
            return lastName1.compareToIgnoreCase(lastName2);
        }
    };
    
    /**
     * Comparing items by number of digits/numbers used after computing frequencies.
     */
    private static class NumbersComparator<T extends Item> implements Comparator<T>
    {
        @Override
        public int compare(T o1, T o2)
        {
            /* Entries computations must be done */
            return ( (new Integer(o1.fetchEntry().numbersUsed())).compareTo((new Integer(o2.fetchEntry().numbersUsed()))) );
        }
    }
    
    /**
     * Get sorting order currently set.
     *
     * @return <code>true</code> when ascending order is active
     */
    public boolean currentSortingOrderAscending()
    {
        return (currentSortOrder == SORT_ORDER.ASCENDING);
    }
    
    /**
     * Sort items according to selected sorting criteria.
     *
     * Sorting order is not changed here.
     *
     * @param sortType type of sorting
     * @param ascending <code>true</code> to set ascending sorting order
     * @see #reverseItems()
     */
    public void sortItems(SORT_TYPE sortType, boolean ascending)
    {
        /* Default comparator */
        Comparator<Item> selectedComparator = chronoComparator;
        
        /* Find out the comparator by criteria */
        if ( sortType == SORT_TYPE.ALPHABETICALLY_BY_FIRSTNAME )
            selectedComparator = alphaComparatorByFirstName;
        else if ( sortType == SORT_TYPE.ALPHABETICALLY_BY_LASTNAME )
            selectedComparator = alphaComparatorByLastName;
        else if ( sortType == SORT_TYPE.BY_NUMBER_FREQS )
            selectedComparator = numbersComparator;
        
        /* Sort */
        Collections.sort(items, selectedComparator);
        
        /* Coudl change sort order */
        if (!ascending)
            Collections.reverse(items);
    }
    
    /**
     * Reverse items inside the collection and toggle current sorting order.
     */
    public void reverseItems()
    {
        Collections.reverse(items);
        currentSortOrder = (currentSortOrder == SORT_ORDER.ASCENDING) ? SORT_ORDER.DESCENDING : SORT_ORDER.ASCENDING;
    }
    
    /* Diagram manipulations */
    
    /**
     * Append diagram into panels.
     *
     * We are looping over panels and then adding the diagram,
     * diagram could be appended into both panels at once.
     *
     * @param item
     * @param insideContentArea     add diagram into main content area
     * @param insideThumbnailArea   add diagram into thumbnail area
     */
    public void addDiagram(Item item, boolean insideContentArea, boolean insideThumbnailArea)
    {
        int t = 2;
        while (t-- > 0)
        {
            boolean updating = (t == 1) ? insideContentArea : insideThumbnailArea;
            if (updating)
            {
                boolean diagramType = (t == 1) ? FULL_DIAGRAM : THUMBNAIL;
                int freePos = window.getFreePosition(diagramType);
                if (freePos < 0)
                    return;

                item.addObserver(this);
                window.setDiagram( getDiagram(item, diagramType), freePos );
            }
        }
    }

    /**
     * Refresh the thumbnail area.
     *
     * @param startItem set this item into the first position, or <code>null</code>
     */
    public void refreshThumbnails(Item startItem)
    {
        int compCount = window.getCurrentNumberOfDiagramPositions(INSIDE_TOP_PANEL);
        int maxThumbnails = window.getMaxNoOfDiagramsInPanel(INSIDE_TOP_PANEL);
        
        /* If nothing could be appended */
        if (compCount == maxThumbnails || (compCount == 0 && startItem == null) )
            return;
        
        /* Updating all the thumbnail area, set starting diagram pivot firstly and then continue */
        if (compCount == 0)
        {
            window.setDiagram( getDiagram(startItem, THUMBNAIL), FIRST_POSITION, INSIDE_TOP_PANEL, INSERT_BEFORE );
            compCount++;
        }
        
        /* Items to append to the end of the panel */
        int index = items.indexOf( ((Item) window.getDiagramOnPos(compCount - 1, INSIDE_TOP_PANEL).fetchEntry()) );
        int moreItems = Math.min( items.size() - (index + 1), maxThumbnails - compCount );
        if ( moreItems > 0 )
        {
            for (int i=1; i<=moreItems; i++)
                addDiagram( items.get(index+i), INSIDE_MAIN_PANEL, INSIDE_TOP_PANEL );
        }
        
        compCount+=moreItems;
        
        if (compCount == maxThumbnails)
            return;
        
        /* Inserting next diagrams if some free positions are at disposal, f.e. when window width changed */
        index = items.indexOf( ((Item) window.getDiagramOnPos(FIRST_POSITION, INSIDE_TOP_PANEL).fetchEntry()) );
        while (index-- > 0 && compCount++ != maxThumbnails)
            window.setDiagram( getDiagram(items.get(index), THUMBNAIL), FIRST_POSITION, THUMBNAIL, INSERT_BEFORE );
    }

    /**
     * Refresh thumbnails when free space should be available inside thumbnail are,
     * useful especially after window maximization.
     */
    public void refreshThumbnails()
    {
        refreshThumbnails(null);
    }
    
    /**
     * Find visible thumbnail and/or full diagram by it's entry.
     *
     * @param search item to search
     * @return found diagrams
     */
    public Diagram[] findDiagrams(Entry search)
    {
        ArrayList<Diagram> foundDiagrams = new ArrayList<Diagram>(2);
        
        int loops = 2;
        while (loops-- > 0)
        {
            boolean isInsideThumbnailArea = (loops == 1) ? false : true;
            int diagramCount = window.getCurrentNumberOfDiagramPositions(isInsideThumbnailArea);
            for (int i=0; i<diagramCount; i++)
            {
                Diagram diagram = window.getDiagramOnPos(i, isInsideThumbnailArea);
                if ( !diagram.isEmpty() && diagram.getEntry().equals(search) )
                    foundDiagrams.add(diagram);
            }
        }
        
        return foundDiagrams.toArray(new Diagram[0]);
    }

    /**
     * Rotating thumbnails.
     *
     * @param count positive values means rotation of count items in the right direction,
     *              else the same count of rotations in the opposite direction
     */
    public void rotateThumbnails(int count)
    {
        /* Counter of current components */
        int cCounter = window.getCurrentNumberOfDiagramPositions(INSIDE_TOP_PANEL);
        
        /* Position on boundary where removing occurs */
        int deletePos = (count > 0) ? FIRST_POSITION : cCounter - 1;
                
        /* Direction of thumbnail rotation */
        boolean rotateToRight = (count > 0) ? true : false;
        
        /* Iterator for items which are going to be diagrams after rotation */
        ListIterator<Item> iterator;
        
        /* Item found by previously declared iterator */
        Item insertItem;
        
        /* Cycles of rotation, direction is remembered by rotateToRight variable */
        count = Math.abs(count);
        
        try
        {
            while (count-- > 0)
            {                                
                iterator = items.listIterator( (rotateToRight ? 1 : 0) + items.indexOf( ((Diagram) window.getDiagramOnPos(cCounter - 1 - deletePos, INSIDE_TOP_PANEL)).fetchEntry() ) );
                insertItem = (rotateToRight) ? iterator.next() : iterator.previous();
                window.removeDiagram(deletePos, INSIDE_TOP_PANEL);
                cCounter--;
                window.setDiagram( getDiagram(insertItem, THUMBNAIL), rotateToRight ? cCounter - deletePos : 0 , INSIDE_TOP_PANEL, !rotateToRight);
            }
        }
        catch (NoSuchElementException e)
        {
            /* No more items to rotate, vector boundaries reached */
        }
    }

    /**
     * Delete all items and diagrams.
     */
    public void deleteAll()
    {
        window.deleteAll();
        items.clear();
        
        /* Set file status modified flag */
        FileManager.getInstance().setFileModifiedFlag(FILE_MODIFIED);
        
        /* No items are selected now */
        selectedItemsCount = 0;
    }
    
    /* Diagram selections */
    
    /**
     * Update number of selected items.
     *
     * @param increment <code>true</code> on selection, <code>false</code> otherwise
     */
    public void updateNoOfItemsSelected(boolean increment)
    {
        if (increment == INCREMENT)
            selectedItemsCount++;
        else
            selectedItemsCount--;
    }
    
    /**
     * Returns number of selected items.
     *
     * @return number of items marked as selected
     */
    public int selectedItems()
    {
        return selectedItemsCount;
    }
    
    /**
     * Deleting of items marked as selected.
     *
     * We delete thumbnails firstly to let user know that something is in progress.
     * Then loop over all registered items and delete the selected items internally.
     * At least the panel with panel should be revalidated.
     */
    public void deleteSelected()
    {
        /* Temporary entry */
        Entry tEntry;
        
        /* Delete all visible thumbnails firstly */
        int thumbnailsCount = window.getCurrentNumberOfDiagramPositions(INSIDE_TOP_PANEL);
        int deletedThumbnails = 0;
        for (int i=0; i<thumbnailsCount; i++)
        {
            tEntry = window.getDiagramOnPos(i - deletedThumbnails, INSIDE_TOP_PANEL).getEntry();
            if ( tEntry.isSelected() )
            {
                window.removeDiagram(i - deletedThumbnails, INSIDE_TOP_PANEL);
                items.remove(tEntry);
                updateNoOfItemsSelected(DECREMENT);
                deletedThumbnails++;
            }
        }
        
        /* Delete everything else selected */
        int itemsCount = getNumberOfItems();
        for (int i=0; i<itemsCount; i++)
        {
            if ( selectedItems() == 0 )
                break;
            
            tEntry = (Entry) items.get(i);
            if ( tEntry.isSelected() )
            {
                items.remove(i--);
                updateNoOfItemsSelected(DECREMENT);
            }
        }

        /* We sometimes need to update top panel after deleting thumbnails 
           and we need to update modified file flag */
        if (deletedThumbnails > 0)
        {
            refreshThumbnails();
            FileManager.getInstance().setFileModifiedFlag(FILE_MODIFIED);
        }
    }

    /* Drag and drop support */
    
    /**
     * Dragging the diagram.
     *
     * @param dge drag event
     */
    public void dragDiagram(DragGestureEvent dge)
    {
        DragAndDropHandler.getInstance().dragDiagram(dge);
    }
    
    /**
     * Dropping the diagram.
     *
     * @param dtde drop event
     */
    public void dropDiagram(DropTargetDropEvent dtde)
    {
        DragAndDropHandler.getInstance().dropDiagram(dtde);
    }

    /* Observer */
    
    /**
     * Updating the observable item.
     *
     * @param observable item or diagram to update
     * @param event type of event
     */
    @Override
    public void update(ItemObservable observable, ItemEvents event)
    {
        if (event == ItemEvents.EVENT_ITEM_SELECTION)
        {
            /* Updating number of selected items on (un)selection */
            updateNoOfItemsSelected( ((Entry) observable).isSelected() );
        }
        else if (event == ItemEvents.EVENT_ENTRY_CHANGED)
        {
            /* The entry (source) string or style changed, so compute characteristics and update
               connected diagrams firstly, then update the item not currently visible */
            if ( observable instanceof Entry )
            {
                final Entry foundEntry = (Entry) observable;
                final String foundEntrySourceString = foundEntry.getSourceString();
                final Style foundEntryStyle = foundEntry.getStyle();
                
                /* Find and update visible diagrams immediatelly */
                final Diagram updatedDiagrams[] = findDiagrams(foundEntry);
                
                for (int i=0; i<updatedDiagrams.length; i++)
                {
                    updatedDiagrams[i].getEntry().setString(foundEntrySourceString);
                    updatedDiagrams[i].getEntry().setStyle(foundEntryStyle);
                    updatedDiagrams[i].forceRepaint();
                    window.updateLabels();
                }
                
                /* Modify entry or diagram if not visible inside panels. */
                final int itemsCount = items.size();
                for (int i=0; i<itemsCount; i++)
                {
                    Entry iEntry = items.get(i).fetchEntry();
                    if ( iEntry.equals(foundEntry) )
                    {
                        iEntry.setString(foundEntrySourceString);
                        iEntry.setStyle(foundEntryStyle);
                    }
                }
            }
        }
    }
}
