package cz.psika.numerist;

import cz.psika.utils.ConfigurationUtils;
import cz.psika.utils.SystemUtils;

import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Class handler for application preferences.
 * 
 * @author Tomas Psika
 */
public class Preferences implements Constants
{
    /* Application preferences */
    
    /* Keys */
    public static final String  PREFERENCE_APPLICATION_VIEW     = "view_mode";
    public static final String  PREFERENCE_DEFAULT_STYLE        = "default_style";
    
    /* Values */
    public static final String  PREFERENCE_DUAL_VIEW            = "dual";
    public static final String  PREFERENCE_SINGLE_VIEW          = "single";
    
    /* Singleton instance */
    private static Preferences prefInstance = null;
    
    /* Stable preferences initialized within application start */
    private Map<String, String> preferences;
    
    /* Preferences change within runtime, will be compared to previous preferences on application exit */
    private Map<String, String> runtimePreferences;
    
    /* Private constructor initializes default preferences */
    private Preferences() 
    {
        preferences = new HashMap<String, String>();
        
        /* Single mode view is default */
        preferences.put(PREFERENCE_APPLICATION_VIEW, PREFERENCE_SINGLE_VIEW);
        
        /* Default style */
        preferences.put(PREFERENCE_DEFAULT_STYLE, defaultStyleName);
    }

    /**
     * Get instance of preferences
     * 
     * Default preferences will be initialized within first call
     * @return single instance of <code>Preferences</code> object
     */
    public static Preferences getInstance()
    {
        if (prefInstance == null)
            return prefInstance = new Preferences();
        
        return prefInstance;
    }
    
    /**
     * Getting preference value
     * 
     * @param key preference key
     * @return value of the preference
     */
    public String getPreference(String key)
    {
        return preferences.get(key);
    }

    /**
     * Loads preferences saved in configuration file
     *
     * When some invalid entries are found, configuration file will be repaired,
     * redundant entries will be ignored by application and appended in case of file repair
     */
    public void fetchUserPreferences()
    {
        String preferencesFile = null;
        Map<String, String> loadedPreferences = new HashMap<String, String>();

        /* Find configuration file, load preferences and check validity against default preferences,
           in case of error default preferences will be used, could potencially check by returned value of readConfiguration() */
        try
        {
            if ( (preferencesFile = ConfigurationUtils.getApplicationConfigFile( SystemUtils.getOSTypeFlag(), APPNAME )) == null )
            {
                /* Configuration could not be created */
                System.err.println("using default preferences");
                return;
            }
            
            ConfigurationUtils.readConfiguration(preferencesFile, loadedPreferences);
        }
        catch (IllegalArgumentException ex)
        {
            System.err.println( ex.getMessage() );
        }
        catch (FileNotFoundException ex)
        {
            /* This should happen very very occasionally */
            System.err.println("could not find application's configuration file, using default preferences");
            return;
        } 
        catch (IOException ex)
        {
            System.err.println("i/o error while loading application's configuratin file, check access permissions, using default preferences instead");
            return;
        }
        
        /* Modify map of preferences and check validity, any uncertainty cause refreshing of configuration file immediatelly ! */
        Iterator<String> iterator = loadedPreferences.keySet().iterator();
        Map<String, String> ignoredPreferences = new HashMap<String, String>();
        int errors = 0;
        String key, value;

        /* Checking all preferences */
        while (iterator.hasNext())
        {
            key = iterator.next();
            value = loadedPreferences.get(key);
            
            /* Redundant values, could be used in later versions */
            if ( ! preferences.containsKey(key) )
            {
                System.err.println("preference '" + key + "' not known and is going to be ignored");
                ignoredPreferences.put(key, value);
                continue;
            }
            
            /* Validity checks */
            if ( key.equalsIgnoreCase(PREFERENCE_APPLICATION_VIEW) )
            {
                /* Single/dual view */
                if ( !value.equalsIgnoreCase(PREFERENCE_SINGLE_VIEW) && !value.equalsIgnoreCase(PREFERENCE_DUAL_VIEW) )
                {
                    System.err.println("preference '" + key +  "' invalid, using single view mode");
                    errors++;
                    continue;
                }
            }
            else if ( key.equalsIgnoreCase(PREFERENCE_DEFAULT_STYLE) )
            {
                /* Any non valid entry will cause fatal error later */
            }
            
            /* Valid preference key and value, modifying defaults */
            preferences.put(key, value);
        }
        
        if (errors > 0)
        {
            /* Need to refresh configuration file because of errors */
            System.err.println("updating configuration file");
            try
            {
                Map<String, String> preferencesRepaired = new HashMap<String, String>();
                preferencesRepaired.putAll(preferences);
                preferencesRepaired.putAll(ignoredPreferences);
                ConfigurationUtils.saveConfiguration(preferencesFile, preferencesRepaired);
                
            } catch (IOException ex)
            {
                System.err.println("updating configuration file failed");
            }
        }
    }
}

