package cz.psika.numerist;

import java.awt.event.ActionEvent;

import java.util.HashMap;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.SwingUtilities;

/**
 * Shareable actions.
 *
 * @author Tomas Psika
 */
public final class SharedActions
{
    /**
     * Singleton dispatcher. 
     */
    private static SharedActions dispatcher;
    
    /**
     * Shareable actions.
     * 
     * @author Tomas Psika
     */
    public enum Actions
    {
        /**
         * Delete all diagrams and update GUI elements.
         */
        DELETE_ALL,
        
        /**
         * Clear diagram(s) inside main content area.
         */
        CLEAR_DIAGRAMS,
        
        /**
         * Saving file contents.
         */
        SAVE_FILE,
        
        /**
         * Delete the diagram selected.
         */
        DELETE_SELECTED,
        
        /**
         * Sort items by currently selected criteria.
         */
        SORT_ITEMS,
        
        /**
         * Change sorting order.
         */
        CHANGE_SORT_TYPE
    }
    
    /**
     * Get singleton instance.
     *
     * @return single instance of this class
     */
    public static final SharedActions getInstance()
    {
        if (dispatcher == null)
            dispatcher = new SharedActions();
        
        return dispatcher;
    }
    
    /**
     * Simple switch returning action handlers without parameters.
     *
     * @param action action type
     * @return action handler
     */
    public static Action getAction(final Actions action)
    {
        if ( action == Actions.DELETE_ALL )
            return deleteAllAction;
        else if (action == Actions.CLEAR_DIAGRAMS)
            return clearDiagramsAction;
        else if (action == Actions.DELETE_SELECTED)
            return deleteSelectedAction;
        else if (action == Actions.CHANGE_SORT_TYPE)
            return changeSortingOrderAction;
        
        throw new IllegalArgumentException("SharedActions.getAction() called with illegal action type");
    }

    /**
     * Get parametrized action handlers.
     *
     * @param action action type
     * @param param action parameter
     * @return action handler
     * @throws IllegalArgumentException when called with action type which is not known
     */
    public Action getAction(final Actions action, final Object param)
    {
        if (action == Actions.SORT_ITEMS)
        {
            if ( !sortingActionsMap.containsKey((Manager.SORT_TYPE) param) )
            {
                SortItemsAction newSortItemsAction = new SortItemsAction();
                newSortItemsAction.setParam( (Manager.SORT_TYPE) param );
                sortingActionsMap.put( (Manager.SORT_TYPE) param, newSortItemsAction );
                return newSortItemsAction;
            }

            return sortingActionsMap.get( (Manager.SORT_TYPE) param );
        }
        
        throw new IllegalArgumentException("SharedActions.getAction()  called with illegal action type with one parameter");
    }
    
    /* Shareable actions */
    
    private static Action deleteAllAction      = new DeleteAllAction();
    private static Action clearDiagramsAction  = new ClearDiagramsAction();
    private static Action deleteSelectedAction = new DeleteSelectedAction();
    
    /**
     * Deleting all diagrams and items.
     */
    private static class DeleteAllAction extends AbstractAction
    {
        @Override
        public void actionPerformed(ActionEvent e)
        {
            Manager.getInstance().deleteAll();
        }
    }

    /**
     * Deleting all selected diagrams and items.
     */
    private static class DeleteSelectedAction extends AbstractAction
    {
        @Override
        public void actionPerformed(ActionEvent e)
        {
            Manager.getInstance().deleteSelected();
        }
    }
    
    /**
     * Erase diagram in content panel.
     */
    private static class ClearDiagramsAction extends AbstractAction
    {
        @Override
        public void actionPerformed(ActionEvent e)
        {
            ApplicationWindow.getInstance().clearDiagrams();
        }
    }

    /**
     * Action which changes sorting order.
     */
    private static Action changeSortingOrderAction = new ChangeSortingOrderAction();
    
    /**
     * Change of sorting order.
     */
    private static class ChangeSortingOrderAction extends AbstractAction
    {
        @Override
        public void actionPerformed(ActionEvent e)
        {
            ApplicationWindow window = ApplicationWindow.getInstance();
            Manager dManager = Manager.getInstance();
            
            /* Reverse items */
            dManager.reverseItems();
            
            /* Clear thumbnails and then append thumbnails */
            Item entry = ApplicationWindow.getInstance().clearThumbnails();
            if (entry != null)
                dManager.refreshThumbnails(entry);
            
            window.updateLabel(Texts.get("sorted"));
        }
    }

    /**
     * Map of all sorting actions.
     */
    private static HashMap<Manager.SORT_TYPE, Action> sortingActionsMap
            = new HashMap<Manager.SORT_TYPE, Action>();
    
    /**
     * Generic sorting action handler.
     */
    private class SortItemsAction extends AbstractAction
    {
        /**
         * Sorting type.
         */
        Manager.SORT_TYPE sortType;
        
        /**
         * Set sorting type of this sorting instance.
         *
         * @param sortTypeRequired
         */
        public void setParam(final Object sortTypeRequired)
        {
            sortType = (Manager.SORT_TYPE) sortTypeRequired;
        }
        
        /**
         * Sorting action.
         *
         * Sorting could be sometimes time-consuming process, so we start new thread to make all necessary updates.
         *
         * @param e event
         */
        @Override
        public void actionPerformed(ActionEvent e)
        {
            final ApplicationWindow window = ApplicationWindow.getInstance();

            /* Sorting algorithm definition */
            final Runnable sorting =  new Runnable() 
                {
                @Override
                    public void run()
                    {
                        window.updateLabel(Texts.get("sorting_..."));
                        final Manager dManager = Manager.getInstance();
                        
                        /* Sort items */
                        dManager.sortItems(sortType, dManager.currentSortingOrderAscending());
                        
                        final Item entry = window.clearThumbnails();
                        if (entry != null)
                            dManager.refreshThumbnails(entry);
                        window.updateLabel(Texts.get("sorted"));
                    }
                };
            
            /* Sorting thread declaration. */
            final Thread sortingThread = new Thread()
                {
                    @Override
                    public void run()
                    {
                        try
                        {
                            SwingUtilities.invokeAndWait(sorting);
                        } catch (Exception e)
                        {
                            System.err.println("sorting error: " + e.getMessage());
                            window.updateLabel(Texts.get("sorting_failed"));
                        }
                    }
                };
            
            /* Start sorting */
            sortingThread.start();
        }
    }
    
    /**
     * Some tests used to make GUI interface behave dynamic.
     *
     * @param action
     * @return <code>true</code> if action could be triggered, else <code>false</code>
     */
    public static boolean actionTest(Actions action)
    {
        if ( action == Actions.DELETE_ALL )
            return ( Manager.getInstance().getNumberOfItems() > 0 ) ? true : false;
        else if ( action == Actions.CLEAR_DIAGRAMS )
            return !ApplicationWindow.getInstance().isContentPanelEmpty();
        else if ( action == Actions.SAVE_FILE )
            return ( FileManager.getInstance().fileModified() && FileManager.getInstance().getCurrentFilename() != null );
        else if ( action == Actions.DELETE_SELECTED )
            return (Manager.getInstance().selectedItems() > 0) ? true : false;
        else if ( action == Actions.SORT_ITEMS )
            return (Manager.getInstance().getNumberOfItems() > 1) ? true : false;
        
        return false;
    }
}
