package cz.psika.numerist;

import cz.psika.utils.ConfigurationUtils;
import java.awt.Color;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Numerology characteristics style settings.
 *
 * @author Tomas Psika
 */
public class Style
{
    /**
     * Map of characters and their corresponding values
     * and some other configuration values.
     */
    protected Map<String, String> map;
    
    /**
     * Mapping of numbers.
     */
    protected Mapping mapping;
    
    /**
     * Default color for digits.
     */
    public static final Color defNumberColor;
    
    /**
     * Default color for lines.
     */
    protected static final Color defLineColor;
    
    /**
     * Default color for diagonal lines.
     */
    protected static final Color defDiagonalLineColor;

    /**
     * Initializing colors.
     */
    static
    {
        defNumberColor = Color.WHITE;
        defLineColor = Color.RED;
        defDiagonalLineColor = Color.GRAY;
    }
    
    /**
     * Name of this style.
     */
    protected String styleName;
    
    /**
     * Protected constructor.
     *
     * @param styleConfigFile file with style definition
     * @exception IOException in case of problems reading configuration file
     * @exception IllegalArgumentException if <code>styleConfigFile</code> is <code>null</code>
     */
    protected Style(final File styleConfigFile) throws IOException
    {
        if (styleConfigFile == null)
            throw new IllegalArgumentException("style file is null");
     
        /* Load style from file and initialize default mapping */
        mapping = new Mapping();
        if ( !loadConfigFromFile(styleConfigFile) )
            throw new IOException("style loading error");
        
        /* Style loaded successfully */
        styleName = styleConfigFile.getName().split(Constants.styleFileSuffix)[0];
        styleName = styleName.substring(0, styleName.length() - 1);
    }
    
    /**
     * Copy constructor.
     * 
     * @param oldStyle old <code>Style</code> object
     */
    public Style(final Style oldStyle)
    {
        styleName = new String( oldStyle.getName() );
        map = Collections.unmodifiableMap( oldStyle.getMap() );
        mapping = oldStyle.getMapping();
    }

    /**
     * Get parameter value of this style.
     * 
     * @param key parameter definition
     * @return value of this style's variable or <code>null</code> when not exists
     */
    public String get(final String key)
    {
       return map.get(key); 
    }
    
    /**
     * Returns digit on given diagram position.
     *
     * @param pos position start with zero, i.e place in top left corner
     * @return number on specified position
     * @see Style.Mapping#numbers
     */
    public int getNumberOnPos(final int pos)
    {
        return mapping.getNumberOnPos(pos);
    }

    /**
     * Returns position of concrete number.
     *
     * @param number or digit, we want to know it's position
     * @return position of specified digit
     * @see Style.Mapping#positions
     */
    public int getPositionOfNumber(final int number)
    {
       return mapping.getPositionOfNumber(number);
    }
    
    /**
     * Each numbers could have it's own color.
     *
     * @param number number in diagram for which we want to know it's color
     * @return color specified by configuration or default number color
     */
    public Color getColorForNumber(final int number)
    {
        return ConfigColor.getColor( map.get("colornumber" + number), defNumberColor );
    }

    /**
     * Returns color of concrete lines.
     * 
     * Ordering from left to right and from top to bottom of diagram.
     *
     * @param line 
     *              <dl>
     *                  <dt>1 .. 3</dt>
     *                      <dd>horizontal lines</dd>
     *                  <dt>4 .. 6</dt>
     *                      <dd>vertical lines</dd> 
     *                  <dt>7 .. 8</dt>
     *                      <dd>inner diagonal lines</dd>
     *                  <dt>9</dt>
     *                      <dd>outer diagonal lines</dd>
     *              </dl>
     * @return assigned line color
     */
    public Color getColorForLine(final int line)
    {
        if (line > 0 && line < 4)
            return ConfigColor.getColor( map.get(Constants.CONFIG_DIAGRAM_ROW_COLOR + line), defLineColor );
        else if (line > 3 && line < 7)
            return ConfigColor.getColor( map.get(Constants.CONFIG_DIAGRAM_COLUMN_COLOR + (line-3) ), defLineColor );
        else if (line > 6 && line < 9)
            return ConfigColor.getColor( map.get(Constants.CONFIG_DIAGRAM_INNER_DIAGONAL_COLOR + (line-6) ), defLineColor );
        else if (line == 9)
            return ConfigColor.getColor( map.get(Constants.CONFIG_DIAGRAM_OUTER_DIAGONAL_COLOR), defDiagonalLineColor);
        
        return defLineColor;
    }
    
    /**
     * Get name of this style.
     *
     * @return this style's name
     */
    public String getName()
    {
        return new String(styleName);
    }
    
    /**
     * Get mapping of numbers <-> numbers.
     * 
     * @return internal object of class <code>Mapping</code>
     */
    public Mapping getMapping()
    {
        return new Mapping( mapping.getSchema() );
    }
    
    /**
     * Get map for this style.
     *
     * @return whole map of style configuration
     */
    private Map<String, String> getMap()
    {
        return new HashMap<String, String>(map);
    }

    /**
     * Loading and parsing of configuration file. Initialization of values.
     *
     * @param styleFile configuration filename
     * @return <code>true</code> on success, <code>false</code> otherwise
     */
    private boolean loadConfigFromFile(final File styleFile)
    {
        try
        {
            /* Initialize configuration map from file */
            map = new HashMap<String, String>(100);
            if ( !ConfigurationUtils.readConfiguration(styleFile.getAbsolutePath(), map) )
                return false;
            
            /* Initialize optional mapping */
            String configMapping;
            if ( ( configMapping = get(Constants.CONFIG_MAPPING_SCHEMA) ) != null )
                mapping = new Mapping(configMapping);
            
            /* Success */
            return true;
        }
        catch (FileNotFoundException ex)
        {
            System.err.println("style file '" + styleFile.getAbsolutePath() + "' not found");
        }
        catch (IOException ex)
        {
            System.err.println("style file '" + styleFile.getAbsolutePath() + "' i/o error");
        }
        catch (Exception ex)
        {
            System.err.println("error while reading style file '" + styleFile.getAbsolutePath() + "'");
        }

        /* Clearing configuration map in case of error */
        map.clear();
        return false;
    }
    
    /**
     * Class representing mapping of numbers and positions in diagram.
     * 
     * @author Tomas Psika
     */
    private class Mapping
    {
        /**
         * Mapping position -> number, array positions are defined as:
         *
         * <table>
         *   <tr>
         *      <td>  0  </td> <td>  1  </td> <td>  2  </td>
         *   </tr>
         *   <tr>
         *      <td>  3  </td> <td>  4  </td> <td>  5  </td>
         *   </tr>
         *   <tr>
         *      <td>  6  </td> <td>  7  </td> <td>  8  </td>
         *   </tr>
         * </table>
         */
        protected int numbers[] = {3,6,9,2,5,8,1,4,7};
        
        /**
         * Positions of concrete numbers inside diagram, 
         * i.e mapping number -> position.
         */
        protected int positions[] = {7,4,1,8,5,2,9,6,3};
        
        /**
         * Current schema, default used on application start.
         */
        protected String currentSchema = defaultSchema;
        
        /**
         * Default mapping schema.
         */
        protected static final String defaultSchema = Constants.CONFIG_MAPPING_DEFAULT;
        
        /**
         * Default constructor initializes default schema.
         */
        private Mapping()
        {
            this(defaultSchema);
        }
        
        /**
         * Mapping constructor.
         *
         * Currently only default mapping schema supported.
         *
         * @param schema type of schema
         */
        private Mapping(final String schema) throws IllegalArgumentException
        {
            /* Mappings are not defined yet */
            if ( !schema.equals(defaultSchema) )
                throw new IllegalArgumentException("unsupported mapping '" + schema + "'");
        }
        
        /**
         * Get mapping schema's name.
         *
         * @return schema type as used in configuration file
         */
        private String getSchema()
        {
            return new String(currentSchema);
        }
        
        /**
         * Get number for defined position in diagram.
         *
         * @param pos position within diagram, zero in top left corner
         * @return number on specicied position
         * @see Style#getNumberOnPos
         */
        protected int getNumberOnPos(final int pos)
        {
            return numbers[pos];
        }
        
        /**
         * Returns position where the number has it's place.
         * 
         * @param number number inside diagram
         * @return position of specified number
         * @see Style#getPositionOfNumber
         */
        protected int getPositionOfNumber(final int number)
        {
            return positions[number];
        }
    }
    
    /**
     * Static class with static methods returning <code>Color</code> objects
     * initialized from configuration strings.
     * 
     * @author Tomas Psika
     */
    protected static class ConfigColor
    {
        /**
         * Get color from configuration string.
         *
         * @param colorString configuration string
         * @param defaultColor default color to return when parsing fails or empty string color supplied via argument
         * @return <code>Color</code> object or default color when not possible to parse string of empty string supplied
         */
        public static Color getColor(final String colorString, final Color defaultColor)
        {
            if (colorString == null)
                return defaultColor;
            
            try
            {
                /* Parse color string */
                String colorStrs[] = colorString.split(",");
                
                return new Color(
                    Integer.valueOf(colorStrs[0]),
                    Integer.valueOf(colorStrs[1]),
                    Integer.valueOf(colorStrs[2])
                    );
            } 
            catch (Exception ex)
            {
                System.err.println("bad color constructor arguments: '" + colorString + "'");
            }
            
            /* Some error occured, returning default color */
            return defaultColor;
        }
    }
}
