package cz.psika.numerist;

import java.io.File;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import cz.psika.utils.ConfigurationUtils;
import cz.psika.utils.SystemUtils;

/**
 * Management of entry styles.
 *
 * @author Tomas Psika
 */
public class StyleManager
{
    /**
     * Current <code>Style</code> object.
     */
    protected static Style currentStyle;
    
    /**
     * Map of all currently usable styles.
     */
    protected static final Map<String, Style> stylesMap;
    
    /**
     * Cache of style files.
     * 
     * There is no need find out regular style files every time,
     * this cache prevents some performance loss.
     */
    protected static final Map<String, File> styleFiles;

    /**
     * Initialize all static class members previously declared.
     */
    static
    {
        stylesMap = new HashMap<String,Style>();
        styleFiles = new HashMap<String, File>();
        currentStyle = null;

        findAllStyles();
    }
    
    /** 
     * Get numerology style of characteristics by name.
     *
     * @param styleName name of style wanted
     * @return <code>Style</code> object or <code>null</code> on error
     */
    public static Style getStyle(final String styleName)
    {
        assert (styleName != null) : "StyleManager.getStyle() argument null";
        
        /* We could already have the style initialized */
        if ( styleLoaded(styleName) )
            return stylesMap.get(styleName);
        
        /* Must load it */
        return loadStyle(styleName);
    }
    
    /**
     * Find out if we have style already in memory.
     *
     * @return <code>true</code> is style was loaded, <code>false</code> otherwise
     */
    private static boolean styleLoaded(final String styleName)
    {
        return stylesMap.containsKey(styleName);
    }

    /**
     * Set current default style of characteristics 
     * used for newly created entries. In case of some error,
     * old settings are rollbacked and we could probably continue.
     *
     * @param styleName name of style
     * @return <code>true</code> on success, <code>false</code> otherwise
     */
    public static boolean setCurrentStyle(final String styleName)
    {
        /* Save previous settings */
        final Style prevStyle = getCurrentStyle();
        
        /* If we used this style previously, return it */
        if ( styleLoaded(styleName) )
        {
            currentStyle = stylesMap.get(styleName);
            return true;
        }
            
        /* Load and check */
        currentStyle = loadStyle(styleName);
        if (currentStyle == null)
        {
            /* Rollback in case of any loading error */
            currentStyle = prevStyle;
            return false;
        }
        else
        {
            /* Successfully loaded style */
            return true;
        }
    }
    
    /**
     * Load current numerology style of characteristics.
     *
     * @return current style or <code>null</code> when not initialized yet
     */
    public static Style getCurrentStyle()
    {
        return currentStyle;
    }
    
    /**
     * Loads style from configuration file.
     *
     * @param loadingStyle  style to load
     * @return found style or <code>null</code>
     */
    protected static Style loadStyle(final String loadingStyle)
    {
        /* Have to load style from configuration file */
        Style newStyle = null;
        try
        {
            newStyle = new Style( styleFiles.get(loadingStyle) );
            stylesMap.put(loadingStyle, newStyle);
        }
        catch (Exception ex)
        {
            System.err.println("style '" +  loadingStyle + "' initialization failed");
            return null;
        }
        
        return newStyle;
    }
    
    /**
     * Get all style names available for now and initialize cache 
     * of style files in order of later use.
     *
     * Standard style definitions could be overloaded, next list shows 
     * priority of directories checked when searching for style files
     * used in application. First items has higher priority. It is possible
     * to e.g. copy style file from /usr/share/numerist/styles directory,
     * copy it to ~/.numerist/styles and change settings. Your settings
     * will be used instead.
     *
     * <p>Directory priorities (first items have higher priority):
     * <ul>
     *      <li>current user's styles</li>
     *      <li>directory specified via JVM '-Dname=value' property on command line</li>
     *      <li>styles in current working directory</li>
     *      <li>shared directory within Linux/Unix standard filesystem hierarchy (supported Unix systems)</li>
     * </ul>
     * </p>
     *
     * @return array of style names
     * @see cz.psika.utils.ConfigurationUtils#getApplicationFiles
     */
    public static String[] findAllStyles()
    {
        /* Clear cache */
        styleFiles.clear();
        stylesMap.clear();
        
        /* Search for configuration files */
        String files[] = 
                ConfigurationUtils.getApplicationFiles(
                    SystemUtils.getOSTypeFlag(), 
                    Constants.APPNAME, 
                    Constants.stylesDirectory, 
                    Constants.styleFileSuffix, 
                    null, 
                    new String[] { System.getProperty("dirstyles") }
                );

        /* Get style names and initialize map used later as cache */
        ArrayList<String> styleNames = new ArrayList<String>();
        for (int t=0; t<files.length; t++)
        {
            String styleName = files[t].substring( files[t].lastIndexOf(File.separator) + 1, files[t].lastIndexOf(".style") );
            if ( ! styleNames.contains(styleName) )
            {
                styleFiles.put(styleName, new File(files[t]));
                styleNames.add(styleName);
            }
        }
        
        return styleNames.toArray(new String[0]);
    }
}
