package cz.psika.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

/**
 * Configuration files seaching and reading
 *
 * Lines beginning with crosshash in configuration file are considered comments,
 * empty lines are ignored.
 * 
 * @author Tomas Psika
 */
public class ConfigurationUtils
{
    private static final String homeDirectory;
    private static final String cwd;
    
    /* Visible or hidden directories/files */
    private static final boolean FILE_VISIBLE    = false;
    private static final boolean FILE_HIDDEN     = true;
    
    static
    {
        /* Finding user's home directory by this property made faithless by JVM */
        homeDirectory = System.getProperty("user.home");
        
        /* Current working directory */
        cwd = System.getProperty("user.dir");
    }
    
    /**
     * Initializing configuration map from file
     *
     * @param file absolute path to config file
     * @param configMap map of configuration values initialized after loading all configuration lines
     * @return <code>true</code> on success, <code>false</code> otherwise
     * @throws FileNotFoundException
     * @throws IOException 
     */
    public static boolean readConfiguration(final String file, final Map<String, String> configMap) throws FileNotFoundException, IOException
    {
        BufferedReader configReader = null;
        String configLine;
        
        try 
        {
            configReader = new BufferedReader(new FileReader(file));
            while ((configLine = configReader.readLine()) != null)
            {
                configLine = configLine.trim();
                int length = configLine.length();
                if ( configLine.startsWith("#") || configLine.length() == 0 )
                    continue;

                int pos = configLine.indexOf("=");
                if (pos < 0)
                {
                    System.err.println("config file format error: \"=\" missing in '" + configLine + "'");
                    configMap.clear();
                    return false;
                }

                configMap.put( configLine.substring(0, pos).trim(), configLine.substring(pos + 1, length).trim() );
            }

            return true;
        } finally 
        {
            if (configReader != null)
                configReader.close();
        }
    }
    
    /**
     * Saving configuration into file
     *
     * @param file file to save
     * @param configMap configuration
     * @throws IOException if configuration file could not be saved
     */
    public static void saveConfiguration(final String file, final Map<String, String> configMap) throws IOException
    {
        BufferedWriter configWriter = null;
        Iterator<String> iterator = configMap.keySet().iterator();
        String configKey;
        
        /* Saving */
        try {
            configWriter = new BufferedWriter( new FileWriter(file) );
            String lineSeparator = System.getProperty("line.separator");
            while ( iterator.hasNext() ) 
            {
                configKey = iterator.next();
                configWriter.write(configKey + "=" + configMap.get(configKey) + lineSeparator);
            }
        }
        finally
        {
            if (configWriter != null)
                configWriter.close();
        }
    }
    
    /**
     * Get configuration files where application data are saved.
     *
     * <p>Configuration files are preferred according to file location 
     * by this preference (first items have higher priority):
     * <ul>
     *      <li>current user's files</li>
     *      <li>directories specified within last argument of type string array</li>
     *      <li>files inside current working directory</li>
     *      <li>files inside user's home directory</li>
     *      <li>shared application directory within Unix standard filesystem hierarchy (on supported Unix systems only)</li>
     * </ul>
     *
     * @param osType type of OS, 'U' means supported Unix system, 'W' means Windows, null none of these
     * @param appName application's name
     * @param subDirectory directory where config files should exist
     * @param suffix searching files with specified suffix only
     * @param filename need to find by exact filename, if non-null suffix will be ignored
     * @param includedDirs directories always needed to append
     * @return array of absolute paths of all relevant files
     * @throws IllegalArgumentException
     */
    public static String[] getApplicationFiles(final String osType, final String appName, final String subDirectory, final String suffix, final String filename, final String[] includedDirs) throws IllegalArgumentException
    {
        String appFilenameHidden, appFilenameVisible;
        try
        {
            appFilenameHidden = getApplicationFilename(osType, appName, FILE_HIDDEN);
            appFilenameVisible = getApplicationFilename(osType, appName, FILE_VISIBLE);
        }
        catch (IllegalArgumentException ex)
        {
            System.err.println("could not get application files");
            return new String[0];
        }

        /* Standard global subdirectories on Unix systems */
        String appDirectoryOnUnix = null;
        if ( osType.equals("U") )
            appDirectoryOnUnix = File.separator + "usr" + File.separator + "share" + File.separator + appFilenameVisible + ( (subDirectory != null) ? File.separator + subDirectory : "" );

        UniqueArrayList<String> dirsForCheck = new UniqueArrayList<String>();
        
        /* User's configurations */
        if ( osType.equals("U") )
            dirsForCheck.add( homeDirectory + File.separator + appFilenameHidden + ( (subDirectory != null) ? File.separator + subDirectory : "" ) );
        else if ( osType.equals("W") )
            dirsForCheck.add( getApplicationDataDirectoryOnWindows() + File.separator + appFilenameHidden + ( (subDirectory != null) ? File.separator + subDirectory : "" ) );
        
        /* Include specified directories */
        if (includedDirs != null)
        {
            for (int i=0; i<includedDirs.length; i++)
                dirsForCheck.add(includedDirs[i]);
        }

        /* Include subdirectories under current working directory */
        if (subDirectory != null)
            dirsForCheck.add(cwd + File.separator + ( (subDirectory != null) ? File.separator + subDirectory : "" ));
        
        /* User's home */
        dirsForCheck.add(homeDirectory);
        
        /* Standard Linux/Unix shareable application directory */
        if (osType.equals("U"))
            dirsForCheck.add(appDirectoryOnUnix);
        
        /* Filter out readable directories only */
        ArrayList<String> dirs = new ArrayList<String>();
        for (int i=0; i<dirsForCheck.size(); i++)
        {
            File fDir = new File( dirsForCheck.get(i) );
            if ( fDir.isDirectory() && fDir.canRead() )
                dirs.add( fDir.getAbsolutePath() );
        }

        /* Filter out concrete files within readable directories */
        File foundFiles[] = new File[0];
        ArrayList<String> foundFilenames = new ArrayList<String>();
        for (int i=0; i<dirs.size(); i++)
        {
            foundFiles = (new File(dirs.get(i))).listFiles( new FilenameFilter() {
                @Override
                public boolean accept(File dir, String name)
                {
                    File configFile = new File(dir + File.separator + name);
                    if (filename != null)
                    {
                        /* Exact match of filename */
                        return (name.equals(filename) && configFile.isFile() && configFile.canRead()) ? true : false;
                    }
                    else
                    {
                        if (suffix != null)
                        {
                            /* Look for the file with specified suffix */
                            return ( name.endsWith("." + suffix) && configFile.isFile() && configFile.canRead() ) ? true : false;
                        }
                        else
                        {
                            /* Every filename matches */
                            return ( configFile.isFile() && configFile.canRead() ) ? true : false;
                        }
                    }
                }
            });
 
            for (int k=0; k<foundFiles.length; k++)
                foundFilenames.add( foundFiles[k].getAbsolutePath() );
        }
        
        return foundFilenames.toArray(new String[0]);
    }
    
    /**
     * Get file with application preferences or create it if none exists
     *
     * @param osType 
     * @param appName 
     * @see #getApplicationFiles
     * @return absolute path to main configuration file or null in case of error
     * @throws IllegalArgumentException
     */
    public static String getApplicationConfigFile(final String osType, final String appName) throws IllegalArgumentException
    {
        final String rcFilename = getApplicationFilename(osType, appName, FILE_HIDDEN).toLowerCase() + "rc";
        
        String configFiles[] = getApplicationFiles(osType, appName, null, null, rcFilename , null );
        if (configFiles.length > 0)
            return configFiles[0];

        /* Configuration file not exists, we will create empty one */
        File configFile = new File(homeDirectory + File.separator + rcFilename);
        try
        {
            if ( !configFile.exists() )
            {
                System.err.println("configuration file does not exist, will be created now");
                configFile.createNewFile();
                return configFile.getAbsolutePath();
            }
        } catch (IOException ex)
        {
            System.err.println("could not create main configuration file");
        }
        
        return null;
    }
    
    /**
     * Get filename of name of directory relevant to this application on specified operating system
     *
     * @param osType OS type flag
     * @param appName application's name
     * @param hidden return directory or file name which should not be visible by default
     * @return relative path to directory or file
     * @throws IllegalArgumentException 
     * @see #getApplicationFiles
     */
    public static String getApplicationFilename(final String osType, String appName, boolean hidden) throws IllegalArgumentException
    {
        if (osType.equals("U"))
        {
            /* Dot prefixed is usual way to make files hidden on most Unix systems */
            return ( (hidden) ? "." : "") + appName.toLowerCase();
        }
        else if (osType.equals("W"))
        {
            /* There is no simple way to make file or directory hidden on Windows systems, so return the same filename */
            return appName.substring(0,1).toUpperCase() + appName.substring(1);
        }
        
        throw new IllegalArgumentException("bad OS type argument");
    }

    /**
     * Try to find out directory where some Windows apps usually save application data
     * without masturbating with Windows API via JNI or something like that (SHGetFolderPath() etc.),
     * on Windows systems first letter of subdirectory name should be uppercased
     */
    private static String getApplicationDataDirectoryOnWindows()
    {
        String appDataDir = System.getenv("APPDATA");
        
        /* When system environment variable is not initialized, use subdirectory within user's home directory */
        if (appDataDir == null)
            appDataDir = homeDirectory;
        
        return appDataDir;
    }
}

